#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import parlai.utils.logging as logging
from projects.seeker.tasks.search_query import WoiSearchQueryTeacher as woi_sgm
from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin
from parlai.tasks.fits.agents import QueryTeacher


class WoiSearchQueryTeacher(BB3TeacherMixin, woi_sgm):
    pass


class FitsSearchQueryTeacher(BB3TeacherMixin, QueryTeacher):
    def __init__(self, opt, shared=None):
        opt['query_source'] = 'human_gold'

        mutators = '+'.join(
            [
                'flatten',
                'prompt_search_query_mutator',
                'skip_retrieval_mutator',
                'fits_pop_keys_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SeeKeR Knowledge Tasks.
"""
import os
import json
from typing import Optional, List

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import MultiTaskTeacher
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.msc.agents as msc
import parlai.tasks.wizard_of_internet.agents as woi
import parlai.tasks.wizard_of_wikipedia.agents as wow
from parlai.tasks.wizard_of_wikipedia.build import build as wow_build_data
import parlai.tasks.squad.agents as squad
import parlai.tasks.ms_marco.agents as ms_marco
import parlai.tasks.triviaqa.agents as triviaqa
import parlai.tasks.natural_questions.agents as nq

import parlai.utils.logging as logging

import projects.seeker.tasks.mutators  # type: ignore


class WoiKnowledgeTeacher(woi.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "WoiKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'flatten',
            'woi_filter_no_passage_used',
            'woi_checked_sentence_as_label',
            'woi_chunk_retrieved_docs',
            'woi_dropout_retrieved_docs',
            'woi_filter_selected_knowledge_in_retrieved_docs',
        ]


class WowKnowledgeTeacher(wow.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['add_missing_turns'] = 'all'
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "WowKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'flatten',
            'wow_filter_no_passage_used',
            'wow_checked_sentence_as_label',
            'wow_to_woi',
            'woi_chunk_retrieved_docs',
            'woi_dropout_retrieved_docs',
            'woi_filter_selected_knowledge_in_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class MsMarcoKnowledgeTeacher(ms_marco.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "MsMarcoKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'ms_marco_filter_has_answer',
            'ms_marco_create_fid_docs',
            'ms_marco_find_selected_sentence_for_knowledge',
            'ms_marco_to_woi',
            'woi_chunk_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class SquadKnowledgeTeacher(squad.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "SquadKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'squad_to_woi',
            'woi_chunk_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class TriviaQAKnowledgeTeacher(triviaqa.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "TriviaQAKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'triviaqa_to_woi',
            'woi_chunk_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class NQKnowledgeTeacher(nq.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "NQKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'nq_to_woi',
            'woi_chunk_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class NQOpenKnowledgeTeacher(nq.NaturalQuestionsOpenTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(self.get_special_mutators())
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = "NQOpenKnowledgeTeacher"

    def get_special_mutators(self) -> List[str]:
        return [
            'nqopen_to_woi',
            'woi_chunk_retrieved_docs',
            'add_selected_sentences_mutator',
        ]


class NQOpenDialoguesKnowledgeTeacher(NQOpenKnowledgeTeacher):
    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id = "NQOpenDialoguesKnowledgeTeacher"

    def setup_data(self, fold):
        data = super().setup_data(fold)
        # Find matching wow dialogue.
        wow_build_data(self.opt)
        wow_filename = os.path.join(
            self.opt['datapath'], 'wizard_of_wikipedia', 'train.json'
        )
        with open(wow_filename, 'r') as f:
            wow_data = json.load(f)
        wow_topics = set([e['chosen_topic'].lower() for e in wow_data])

        def try_find_matching_topic(q):
            for topic in wow_topics:
                if (
                    f' {topic.lower()} ' in q.lower()
                    or f' {topic.lower()} ' in q.lower()
                ):
                    return topic
            return None

        def context_for_topic(topic):
            context = []
            for ex in wow_data:
                if ex['chosen_topic'].lower() == topic:
                    context = [d['text'] for d in ex['dialog']]
                    break
            # Make sure that the dialogue doesn't end on a question.
            while context and context[-1].strip().endswith('?'):
                context.pop()
            return context

        for ex, done in data:
            question = ex['text']
            topic = try_find_matching_topic(question)
            if not topic:
                continue
            context = context_for_topic(topic)
            if not context:
                continue
            ex['history'] = '\n'.join(context)
            yield ex, done


def get_dialogue_task_mutators(opt: Opt) -> str:
    """
    Set the mutators appropriately for the dialogue tasks.
    """
    mutators = '+'.join(
        ['flatten', 'extract_entity_for_knowledge_model', 'skip_retrieval_mutator']
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class Convai2KnowledgeTeacher(convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = self.get_special_mutators(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)
        self.id = 'Convai2KnowledgeTeacher'

    def get_special_mutators(self, opt):
        return get_dialogue_task_mutators(opt)


class EDKnowledgeTeacher(ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = self.get_special_mutators(opt)
        super().__init__(opt, shared)
        self.id = 'EDKnowledgeTeacher'

    def get_special_mutators(self, opt):
        return get_dialogue_task_mutators(opt)


class BSTKnowledgeTeacher(bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = self.get_special_mutators(opt)
        super().__init__(opt, shared)
        self.id = 'BSTKnowledgeTeacher'

    def get_special_mutators(self, opt):
        return get_dialogue_task_mutators(opt)


class MSCKnowledgeTeacher(msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = self.get_special_mutators(opt)
        opt['include_session1'] = False
        super().__init__(opt, shared)
        self.id = 'MSCKnowledgeTeacher'

    def get_special_mutators(self, opt):
        return get_dialogue_task_mutators(opt)


class MSCKnowledgeOverlapTeacher(msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'msc_find_selected_sentence_knowledge',
                'add_retrieved_documents_mutator',
                'skip_retrieval_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['include_session1'] = False
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'MSCKnowledgeOverlapTeacher'


class KnowledgeTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        WoiKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        WowKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        MsMarcoKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        SquadKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        TriviaQAKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        NQKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        NQOpenKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        NQOpenDialoguesKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        Convai2KnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        EDKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        BSTKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        MSCKnowledgeTeacher.add_cmdline_args(parser, partial_opt)
        MSCKnowledgeOverlapTeacher.add_cmdline_args(parser, partial_opt)
        return parser

    def __init__(self, opt, shared=None):
        tasks = [
            f"projects.seeker.tasks.knowledge:{teacher}"
            for teacher in [
                'WoiKnowledgeTeacher',
                'WowKnowledgeTeacher',
                'MsMarcoKnowledgeTeacher',
                'SquadKnowledgeTeacher',
                'TriviaQAKnowledgeTeacher',
                'NQKnowledgeTeacher',
                'NQOpenKnowledgeTeacher',
                'NQOpenDialoguesKnowledgeTeacher',
                'Convai2KnowledgeTeacher',
                'EDKnowledgeTeacher',
                'BSTKnowledgeTeacher',
                'MSCKnowledgeTeacher',
                'MSCKnowledgeOverlapTeacher',
            ]
        ]
        opt['task'] = ','.join(tasks)
        super().__init__(opt, shared)


class DefaultTeacher(KnowledgeTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Any, Dict, List, Optional, Union

from parlai.agents.rag.retrieve_api import (
    SearchEngineRetriever,
    SearchEngineRetrieverMock,
)
from parlai.agents.rag.retrievers import Document
from parlai.core.agents import Agent
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.message import Message


SERVERS = {
    'default': 'RELEVANT_SEARCH_SERVER',
    'test': 'http://test_api',
}


class SearchAgent(Agent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            '--server',
            type=str,
            choices=SERVERS.keys(),
            default='default',
            help='Which search server to use',
        )
        parser.add_argument(
            '--raw-server',
            type=str,
            default=None,
            help='Specify to override the server choices with your own.',
        )
        parser.add_argument(
            '--n-docs', type=int, default=5, help='How many docs to retrieve'
        )
        parser.add_argument(
            '--intra-doc-delimiter',
            type=str,
            default='\n',
            help='How to delimit intra-document contents',
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared=shared)
        opt['search_server'] = SERVERS[opt['server']]
        if opt.get('raw_server') is not None:
            opt['search_server'] = opt['raw_server']
        opt['skip_retrieval_token'] = ''
        self.intra_doc_delim = opt['intra_doc_delimiter']
        self.n_docs = opt['n_docs']
        if shared is None:
            self.search_client = (
                SearchEngineRetriever(opt)
                if opt['server'] != 'test'
                else SearchEngineRetrieverMock(opt)
            )
        else:
            self.search_client = shared['client']
        self.top_docs = []

    def reset(self):
        super().reset()
        self.top_docs = []

    def share(self) -> Dict[str, Any]:
        shared = super().share()
        shared['client'] = self.search_client
        return shared

    def act(self):
        observation = self.observation
        results = self.search_client.retrieve([observation['text']], self.n_docs)[0]
        documents: List[Document] = []
        for doc in results:
            content = (
                self.intra_doc_delim.join(doc['content'])
                if isinstance(doc['content'], list)
                else doc['content']
            )
            documents.append(
                Document(docid=doc['url'], text=content, title=doc['title'])
            )
        reply = {
            'text': '\n'.join([str(doc) for doc in documents]),
            'top_docs': documents,
        }
        return reply

    def respond(
        self, text_or_message: Union[str, Message], **other_message_fields
    ) -> str:
        """
        Override Agent.respond to set top_docs.
        """
        if isinstance(text_or_message, str):
            observation = Message(text=text_or_message, **other_message_fields)
        else:
            observation = Message(**text_or_message, **other_message_fields)
            if 'text' not in observation:
                raise RuntimeError('The agent needs a \'text\' field in the message.')

        if 'episode_done' not in observation:
            observation['episode_done'] = True
        agent = self.clone()
        agent.observe(observation)
        response = agent.act()
        self.top_docs = response['top_docs']
        return response['text']



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Modules for SeeKeR.
"""
import torch
from typing import List, Tuple, Optional, Dict, Any

from parlai.agents.fid.fid import FidModel
from parlai.agents.rag.args import RetrieverType
from parlai.agents.rag.retrievers import (
    Document,
    BLANK_DOC,
    SearchQuerySearchEngineRetriever,
    RagRetriever,
    retriever_factory,
)
from parlai.core.dict import DictionaryAgent
from parlai.core.opt import Opt
import parlai.utils.logging as logging


def interleave_fid_combo_outputs(
    enc_out_retrieval: torch.Tensor,
    enc_out_skip_retrieval: torch.Tensor,
    mask_retrieval: torch.Tensor,
    mask_skip_retrieval: torch.Tensor,
    skip_retrieval_vec: torch.Tensor,
    top_docs: List[List[Document]],
    top_doc_scores: torch.Tensor,
    right_padded: bool = True,
) -> Tuple[torch.Tensor, torch.BoolTensor, List[List[Document]], torch.Tensor]:
    """
    Interleave FiD encoder outputs.

    Outputs are either encodings of documents/input,
    or encodings of just input.

    This operation is to preserve original batch order.

    :param enc_out_retrieval:
        encoder output for inputs with retrieved documents
    :param enc_out_skip_retrieval:
        encoder output for inputs without retrieved documents
    :param mask_retrieval:
        mask for inputs with retrieved documents
    :param mask_skip_retrieval:
        mask for inputs without retrieved documents
    :param skip_retrieval_vec:
        1D tensor indicating which inputs skip retrieval vs. dont
    :param top_docs:
        top documents for outputs with retrieved documents
    :param top_doc_scores:
        top doc scores.
    :param right_padded:
        whether the final output is right padded

    :return (new_out, new_mask, new_top_docs, new_top_doc_scores):
        return interleaved concatenation of inputs
    """
    bsz = enc_out_retrieval.size(0) + enc_out_skip_retrieval.size(0)
    dim = enc_out_retrieval.size(-1)
    seqlen = max(enc_out_retrieval.size(1), enc_out_skip_retrieval.size(1))
    n_docs = top_doc_scores.size(1)

    # The following logic interleaves the results in new tensors.
    new_out = enc_out_retrieval.new(bsz, seqlen, dim).fill_(0)
    new_mask = mask_retrieval.new(bsz, seqlen).fill_(False)
    new_top_doc_scores = top_doc_scores.new(bsz, n_docs)
    new_top_docs = []
    retr_offset = 0
    skip_offset = 0
    for i, skip in enumerate(skip_retrieval_vec):
        if skip:
            vec = enc_out_skip_retrieval[skip_offset]
            mask = mask_skip_retrieval[skip_offset]
            scores = torch.zeros(n_docs).to(top_doc_scores)
            docs = [BLANK_DOC] * n_docs
            skip_offset += 1
        else:
            vec = enc_out_retrieval[retr_offset]
            mask = mask_retrieval[retr_offset]
            scores = top_doc_scores[retr_offset]
            docs = top_docs[retr_offset]
            retr_offset += 1
        if right_padded:
            new_out[i, : vec.size(0), :] = vec
            new_mask[i, : mask.size(0)] = mask
        else:
            new_out[i, -vec.size(0) :, :] = vec
            new_mask[i, -mask.size(0) :] = mask

        new_top_doc_scores[i] = scores
        new_top_docs.append(docs)

    return new_out, new_mask, new_top_docs, new_top_doc_scores  # type: ignore


class ComboFidSearchRetrieverMixin:
    """
    Mixin for Combo Search Modules.

    These models **either** search **or** do not.
    """

    def set_search_queries(self, queries: List[str]):
        """
        Bypass SQ generation by setting the search queries.
        """
        self.search_queries = queries

    def init_search_query_generator(self, opt):
        """
        If a generator model file is specified, we use it.
        """
        if opt['search_query_generator_model_file']:
            return super().init_search_query_generator(opt)  # type: ignore
        else:
            logging.warning('Not initializing SQ Generator *within* retriever')

    def generate_search_query(self, query: torch.LongTensor) -> List[str]:
        """
        Generate search queries.

        If there are cached queries, use those; otherwise, generate our own.
        """
        if self.search_queries:
            # Ensure we only search for examples with well-formed queries
            # as this list is bsz-length, not search-bsz-length.
            search_queries = [q for q in self.search_queries if q]
        else:
            search_queries = super().generate_search_query(query)  # type: ignore
        return search_queries


class ComboFidSearchQuerySearchEngineRetriever(
    ComboFidSearchRetrieverMixin, SearchQuerySearchEngineRetriever
):
    """
    Override Search Engine Retriever to accommodate SQ Generator from Combo Setup.
    """


def combo_fid_retriever_factory(
    opt: Opt, dictionary: DictionaryAgent, shared=None
) -> Optional[RagRetriever]:
    """
    Bypass call to standard retriever factory to possibly build our own retriever.
    """
    if opt.get('converting'):
        return None
    retriever = RetrieverType(opt['rag_retriever_type'])
    if retriever is RetrieverType.SEARCH_ENGINE:
        return ComboFidSearchQuerySearchEngineRetriever(
            opt, dictionary, shared=shared
        )  # type: ignore
    else:
        return retriever_factory(opt, dictionary, shared)


class ComboFidModel(FidModel):
    """
    Combo FiD Model.

    This model receives a new input that indicates which batch items require
    retrieval/FiD encoding, and which can be used in a standard encoder/decoder pass.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, retriever_shared=None):
        super().__init__(opt, dictionary, retriever_shared)
        self.top_docs = []

    @classmethod
    def build_retriever(
        cls,
        opt: Opt,
        dictionary: DictionaryAgent,
        retriever_shared: Optional[Dict[str, Any]],
    ) -> Optional[RagRetriever]:
        return combo_fid_retriever_factory(opt, dictionary, retriever_shared)

    def set_search_queries(self, queries: List[str]):
        """
        Set retriever's search queries.
        """
        assert self.retriever is not None
        self.retriever.set_search_queries(queries)

    def get_top_docs(self) -> List[List[Document]]:
        """
        Return Documents.
        """
        return self.top_docs

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        skip_retrieval_vec: torch.BoolTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        """
        Override Standard FiD.encoder to account for skip_retrieval inputs.

        With `skip_retrieval` inputs, we forward directly through the `seq2seq_encoder`

        Otherwise, we call super().encoder to retrieve and concatenate outputs.

        :param input:
            2D [bsz, seqlen] input to the encoder
        :param input_lengths:
            1D [bsz] lengths of each input item
        :param query_vec:
            2D [bsz*n_turns, seqlen] input for the retriever
        :param input_turns_cnt:
            1D [bsz] number of dialogue turns for each input example
        :param skip_retrieval_vec:
            1D [bsz] indicator for whether to skip retrieval.

        :return (encoder_out, encoder_mask, input_turns_cnt, top_docs, top_doc_scores):
            encoder_out: *concatenated* (or not) encoded representations of context/document pairs
            encoder_mask: new mask for enc_out
            input_turns_cnt: pass along the input turns count for the decoder
            top_docs: List of top Documents for each batch example
            top_doc_scores: scores for each retrieved document.
        """
        # If we have a full batch of retrieve (or not), can use simple logic here.
        if torch.all(skip_retrieval_vec):
            new_out, new_mask = self.seq2seq_encoder(input, positions, segments)
            return new_out, new_mask, None, None, None
        elif torch.all(~skip_retrieval_vec):
            output = super().encoder(
                input, input_lengths, query_vec, input_turns_cnt, positions, segments
            )
            self.top_docs = output[-2]
            return output

        assert all(t is None for t in [input_turns_cnt, positions, segments])
        # Encode with `super()` call for non-skip-retrieval inputs
        (
            enc_out_retrieval,
            mask_retrieval,
            input_turns_cnt,
            top_docs,
            top_doc_scores,
        ) = super(ComboFidModel, self).encoder(
            input[~skip_retrieval_vec],
            input_lengths[~skip_retrieval_vec],
            query_vec[~skip_retrieval_vec],
            input_turns_cnt,
            positions,
            segments,
        )
        # Encode with seq2seq_encoder for skip-retrieval inputs
        enc_out_skip_retrieval, mask_skip_retrieval = self.seq2seq_encoder(
            input[skip_retrieval_vec]
        )

        (
            new_out,
            new_mask,
            new_top_docs,
            new_top_doc_scores,
        ) = interleave_fid_combo_outputs(
            enc_out_retrieval,
            enc_out_skip_retrieval,
            mask_retrieval,
            mask_skip_retrieval,
            skip_retrieval_vec,
            top_docs,
            top_doc_scores,
        )
        self.top_docs = new_top_docs

        return new_out, new_mask, input_turns_cnt, new_top_docs, new_top_doc_scores



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest  # noqa: F401


class TestFitsBaseTeacher(AutoTeacherTest):
    task = 'fits'


class TestFitsFeedbackTeacher(AutoTeacherTest):
    task = 'fits:FitsFeedbackTeacher'


class TestFitsSatisfactionTeacher(AutoTeacherTest):
    task = 'fits:FitsSatisfactionTeacher'


class TestQueryTeacher(AutoTeacherTest):
    task = 'fits:QueryTeacher'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BB3 Module Definition.
"""
from enum import Enum
from typing import Dict, List

import projects.bb3.constants as CONST


class Module(Enum):

    SEARCH_DECISION = 'sdm'
    MEMORY_DECISION = 'mdm'
    SEARCH_QUERY = 'sgm'
    MEMORY_GENERATOR = 'mgm'
    CONTEXTUAL_KNOWLEDGE = 'ckm'
    MEMORY_KNOWLEDGE = 'mkm'
    SEARCH_KNOWLEDGE = 'skm'
    CONTEXTUAL_DIALOGUE = 'crm'
    MEMORY_DIALOGUE = 'mrm'
    SEARCH_DIALOGUE = 'srm'
    VANILLA_DIALOGUE = 'vrm'
    GROUNDED_DIALOGUE = 'grm'
    OPENING_DIALOGUE = 'orm'
    BOT_DIALOGUE = 'brm'

    @staticmethod
    def dialogue_modules() -> List['Module']:
        return [
            Module.CONTEXTUAL_DIALOGUE,
            Module.MEMORY_DIALOGUE,
            Module.SEARCH_DIALOGUE,
            Module.VANILLA_DIALOGUE,
            Module.GROUNDED_DIALOGUE,
            Module.BOT_DIALOGUE,
            Module.OPENING_DIALOGUE,
        ]

    @staticmethod
    def knowledge_modules() -> List['Module']:
        return [
            Module.CONTEXTUAL_KNOWLEDGE,
            Module.MEMORY_KNOWLEDGE,
            Module.SEARCH_KNOWLEDGE,
        ]

    @staticmethod
    def decision_modules() -> List['Module']:
        return [
            Module.SEARCH_DECISION,
            Module.MEMORY_DECISION,
        ]

    def decision_do_key(self) -> str:
        return {
            Module.SEARCH_DECISION: 'search_decision_do_search_reply',
            Module.MEMORY_DECISION: 'memory_decision_do_access_reply',
        }[self]

    def decision_dont_key(self) -> str:
        return {
            Module.SEARCH_DECISION: 'search_decision_dont_search_reply',
            Module.MEMORY_DECISION: 'memory_decision_dont_access_reply',
        }[self]

    def message_name(self) -> str:
        """
        Name used to access output of a module in a Message.
        """
        return self.tag_to_agent()[self.value]

    def agent_name(self):
        """
        Display name for user's, and debugging, sake.
        """
        return f"{self.tag_to_agent()[self.value]}_agent"

    def model_file_path_key(self):
        """
        Opt key for model file path for this agent.
        """
        return f"{self.tag_to_agent()[self.value]}_response_model_path"

    def tag(self):
        return self.value

    def is_dialogue(self):
        return self in Module.dialogue_modules()

    def is_knowledge(self):
        return self in Module.knowledge_modules()

    def skip_search(self):
        return self.value not in ['mkm', 'skm']

    def is_one_turn_history(self):
        return self.value in ['mdm', 'sdm', 'mgm']

    @staticmethod
    def tag_to_agent() -> Dict[str, str]:
        return {
            'sdm': 'search_decision',
            'mdm': 'memory_decision',
            'sgm': 'search_query',
            'mgm': 'memory_generator',
            'ckm': 'contextual_knowledge',
            'mkm': 'memory_knowledge',
            'skm': 'search_knowledge',
            'crm': 'contextual_dialogue',
            'mrm': 'memory_dialogue',
            'srm': 'search_dialogue',
            'vrm': 'vanilla_dialogue',
            'grm': 'grounded_dialogue',
            'orm': 'opening_dialogue',
            'brm': 'bot_dialogue',
        }

    ##############
    # R2C2 Func. #
    ##############
    def r2c2_prompt(self):
        """
        Prompt token for this module.
        """
        return {
            'sdm': CONST.IS_SEARCH_REQUIRED,
            'mdm': CONST.IS_MEMORY_REQUIRED,
            'sgm': CONST.GENERATE_QUERY,
            'mgm': CONST.GENERATE_MEMORY,
            'mkm': CONST.ACCESS_MEMORY,
            'ckm': CONST.EXTRACT_ENTITY,
            'skm': CONST.GENERATE_KNOWLEDGE,
            'mrm': '',
            'crm': '',
            'srm': '',
            'vrm': '',
            'grm': '',
            'orm': '',
            'brm': '',
        }[self.value]

    def special_tokens(self):
        return {
            Module.CONTEXTUAL_KNOWLEDGE.message_name(): (
                CONST.BEGIN_ENTITY,
                CONST.END_ENTITY,
            ),
            Module.MEMORY_KNOWLEDGE.message_name(): (
                CONST.BEGIN_MEMORY,
                CONST.END_MEMORY,
            ),
            Module.SEARCH_KNOWLEDGE.message_name(): (
                CONST.TOKEN_KNOWLEDGE,
                CONST.TOKEN_END_KNOWLEDGE,
            ),
        }[self.tag_to_agent()[self.value]]

    #############
    # OPT Func. #
    #############
    def opt_prompt(self):
        """
        Prompt token for OPT models.
        """
        return {
            'sdm': "Person 2 must decide whether to search the internet.\n\n",
            'mdm': "A conversation between two persons. Person 2 must consult their notes about Person 1.\n\n",
            'sgm': "Person 2 must write a search query for a search engine.\n\n",
            'mgm': "A conversation between two persons. Person 2 writes a note about Person 1 to help remember information for later.\n\n",
            'ckm': "A conversation between two persons. Person 2 recalls a previous topic in the conversation.\n\n",
            'skm': "A conversation between two persons. Person 2 finds an interesting fact from the internet.\n\n",
            'mkm': "A conversation between two persons. Person 2 recalls an interesting fact about Person 1 or Person 2.\n\n",
            'crm': "A conversation between two persons. Person 2 would like to continue talking about a previous topic in the conversation.\n\n",
            'mrm': "A conversation between two persons. Person 2 would like to chat about an interesting fact about Person 1 or Person 2.\n\n",
            'srm': "A conversation between two persons. Person 2 would like to tell Person 1 about something Person 2 found on the internet.\n\n",
            'vrm': "A conversation between two persons.\n\n",
            'grm': "A conversation between two persons. Person 2 responds in a given style.\n\n",
            'orm': "A conversation between two persons. Person 2 begins the conversation given information about Person 1.\n\n",
            'brm': "A conversation between two persons.\n\n",
        }[self.value]

    def opt_final_prefix(self):
        """
        Final prefix to put after constructing context for OPT.
        """
        import projects.bb3.prompts as PROMPT

        return {
            'sdm': PROMPT.SEARCH_DECISION,
            'mdm': PROMPT.MEMORY_DECISION,
            'sgm': PROMPT.QUERY_GEN_PREFIX,
            'mgm': PROMPT.MEMORY_GEN_PREFIX,
            'mkm': PROMPT.MEMORY_KNOWLEDGE_PREFIX,
            'ckm': PROMPT.CONTEXTUAL_KNOWLEDGE_PREFIX,
            'skm': PROMPT.SEARCH_KNOWLEDGE_PREFIX,
            'mrm': PROMPT.SELF_PREFIX,
            'crm': PROMPT.SELF_PREFIX,
            'srm': PROMPT.SELF_PREFIX,
            'vrm': PROMPT.SELF_PREFIX,
            'grm': PROMPT.SELF_PREFIX,
            'orm': PROMPT.OPENING_PREFIX,
            'brm': PROMPT.SELF_PREFIX,
        }[self.value]

    def opt_shots(self) -> str:
        import projects.bb3.prompts as PROMPT

        return PROMPT.SHOTS[self]

    def opt_pre_context_tok(self):
        import projects.bb3.prompts as PROMPT

        if (
            (self.is_knowledge() and self is not Module.CONTEXTUAL_KNOWLEDGE)
            or self is Module.GROUNDED_DIALOGUE
            or self is Module.BOT_DIALOGUE
        ):
            return PROMPT.PRE_CONTEXT_TOK
        return ''

    def opt_post_context_tok(self):
        import projects.bb3.prompts as PROMPT

        if self.is_dialogue() and self not in [
            Module.VANILLA_DIALOGUE,
            Module.OPENING_DIALOGUE,
            Module.GROUNDED_DIALOGUE,
            Module.BOT_DIALOGUE,
        ]:
            return PROMPT.POST_CONTEXT_TOK
        return ''

    def opt_dialogue_knowledge_prefix(self):
        return {
            'mrm': "Personal Fact: ",
            'crm': "Previous Topic: ",
            'srm': "Interesting Fact: ",
        }[self.value]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from datetime import datetime
from typing import Any, Dict

import numpy as np
import pandas as pd

from parlai.crowdsourcing.utils.acceptability import AcceptabilityChecker
from parlai.crowdsourcing.utils.analysis import AbstractTurnAnnotationResultsCompiler
from parlai.crowdsourcing.tasks.model_chat.model_chat_blueprint import BLUEPRINT_TYPE

# importing BLUEPRINT_TYPE to force registration of the blueprint, not using this var itself
_ = BLUEPRINT_TYPE


class ModelChatResultsCompiler(AbstractTurnAnnotationResultsCompiler):
    """
    Compile and save results of human+model chats.

    Results will be saved on the level of specific conversations, as well as aggregated
    up the level of each worker as a whole.
    """

    @classmethod
    def setup_args(cls):
        parser = super().setup_args()
        parser.add_argument(
            '--max-convos-per-worker',
            type=int,
            default=100,
            help='The most conversations to analyze from any one user. Set to -1 for no limit.',
        )
        parser.add_argument(
            '--min-word-count',
            type=int,
            default=4,
            help='The minimum acceptable mean number of words per human utterance',
        )
        parser.add_argument(
            '--worker-block-list',
            type=str,
            default='',
            help='Comma-separated list of all workers to block',
        )
        return parser

    def __init__(self, opt: Dict[str, Any]):

        super().__init__(opt)
        # Validate problem buckets
        if self.use_problem_buckets and 'none_all_good' not in self.problem_buckets:
            # The code relies on a catchall "none" category if the user selects no other
            # annotation bucket
            raise ValueError(
                'There must be a "none_all_good" category in self.problem_buckets!'
            )

        # Input args
        os.makedirs(self.output_folder, exist_ok=True)
        self.max_convos_per_worker = opt['max_convos_per_worker']
        self.min_word_count = opt['min_word_count']
        self.worker_block_list = opt['worker_block_list'].split(',')

        # Setting up problem buckets
        if self.use_problem_buckets:
            self.regular_buckets = [
                bucket
                for bucket in self.problem_buckets
                if bucket not in ['other', 'none_all_good']
            ]
            # Remove the buckets that are special cases

        self.acceptability_checker = AcceptabilityChecker()

    def get_results_path_base(self) -> str:
        now = datetime.now()
        return os.path.join(
            self.output_folder, f'results_{now.strftime("%Y%m%d_%H%M%S")}'
        )

    def compile_results(self) -> pd.DataFrame:
        task_units_data = self.get_task_data()
        now = datetime.now()
        worker_results_file = os.path.join(
            self.output_folder, f'worker_results_{now.strftime("%Y%m%d_%H%M%S")}.csv'
        )
        # Read in each file
        num_convos_with_no_save_data = 0
        num_wrong_status_convos = 0
        num_complete_convos = 0
        complete_convos_per_model = {}
        bad_conversations = []
        stat_counts = {}
        worker_stats = {}
        worker_conversation_counts = {}
        total_utterances = 0

        conversation_idx = 0
        conversation_dfs = []
        for task_unit in task_units_data:

            worker_id = task_unit['worker_id']
            assignment_id = task_unit['assignment_id']

            # Determining whether the task unit should be skipped
            # Extract out custom data
            if task_unit['data']['save_data'] is None:
                num_convos_with_no_save_data += 1
                continue
            elif task_unit['status'] not in ['completed', 'approved']:
                num_wrong_status_convos += 1
                continue
            else:
                num_complete_convos += 1

            # Read in file
            data = task_unit['data']['save_data']['custom_data']

            # Only include the first max_convos_per_worker conversations from a
            # worker to avoid biasing
            worker_id = task_unit['worker_id']
            assignment_id = task_unit['assignment_id']
            if worker_id in worker_conversation_counts:
                conversations_so_far = worker_conversation_counts[worker_id]
            else:
                conversations_so_far = 0
            worker_conversation_counts[worker_id] = conversations_so_far + 1
            if (
                self.max_convos_per_worker != -1
                and conversations_so_far >= self.max_convos_per_worker
            ):
                print(
                    f'Had {conversations_so_far} conversation(s) already from this worker {worker_id}. Skipping {assignment_id}.'
                )
                continue

            # Check if need to block the turker
            word_counts = [
                len(d['text'].split(' ')) for d in data['dialog'] if d['agent_idx'] == 0
            ]
            utterances = [d['text'] for d in data['dialog'] if d['agent_idx'] == 0]
            if np.average(word_counts) < self.min_word_count:
                bad_conversations.append(data)
                print(
                    f'Bad complete conversation, words from human: {utterances}. Skipping.'
                )
                continue

            if self.use_problem_buckets:
                if not all(
                    bucket in data['dialog'][1]['problem_data']
                    for bucket in self.problem_buckets
                ):
                    raise ValueError('Bucket(s) are missing from the problem data!')

            model_nickname = data['task_description']['model_nickname']
            if model_nickname not in stat_counts:
                stat_counts[model_nickname] = {}
            if model_nickname in complete_convos_per_model:
                complete_convos_per_model[model_nickname] += 1
            else:
                complete_convos_per_model[model_nickname] = 1

            # Extract non-message info
            info_dict = {
                'worker': worker_id,
                'model_nickname': model_nickname,
                'bad_workers': ','.join(data['bad_workers']),
                'hit_id': data['hit_ids'][0],
                'assignment_id': assignment_id,
                'context_dataset': data['context_dataset'],
                'additional_context': data['additional_context'],
            }

            # Check that the conversation consists of pairs of comments between
            # agents 0 and 1, with 0 speaking first
            assert all(
                [
                    utterance_data['agent_idx'] == utterance_idx % 2
                    for utterance_idx, utterance_data in enumerate(data['dialog'])
                ]
            )

            # Determine whether the HIT contains unacceptable messages.
            # (We do this for every HIT, even if acceptability violation info
            # was already saved, because the violation criteria may have
            # changed since the HIT was collected.)
            messages_0 = [utt for utt in data['dialog'] if utt['agent_idx'] == 0]
            messages_1 = [utt for utt in data['dialog'] if utt['agent_idx'] == 1]
            assert len(messages_0) + len(messages_1) == len(data['dialog'])

            # Check the human utterances for safety
            utterances_0 = [m['text'] for m in messages_0]
            info_dict[
                'acceptability_violations_0'
            ] = self.acceptability_checker.check_messages(
                messages=utterances_0,
                is_worker_0=True,
                violation_types=self.acceptability_checker.ALL_VIOLATION_TYPES,
            )

            # Compile personas and previous utterances
            df = pd.DataFrame(
                [],
                columns=[
                    'worker_id',
                    'hit_id',
                    'model_nickname',
                    'conversation_idx',
                    'turn_idx',
                    'agent_idx',
                    'text',
                ]
                + self.problem_buckets,
            )
            text_parts = []
            if data['personas'] is not None and len(data['personas']) > 0:
                text_parts += [
                    'your persona: ' + data['personas'][1][0],
                    'your persona: ' + data['personas'][1][1],
                ]
            if (
                data['additional_context'] is not None
                and len(data['additional_context']) > 0
            ):
                text_parts.append(data['additional_context'])
            df = df.append(
                {
                    'worker_id': info_dict['worker'],
                    'hit_id': info_dict['hit_id'],
                    'model_nickname': model_nickname,
                    'conversation_idx': conversation_idx,
                    'turn_idx': -1,
                    'agent_idx': 1,
                    'text': '\n'.join(text_parts),
                    **{bucket: '' for bucket in self.problem_buckets},
                },
                ignore_index=True,
            )

            total_utterances += len([d for d in data["dialog"] if d["agent_idx"] == 1])
            if len(data['dialog']) > 20:
                print(
                    f'Got long dialogue of {len(data["dialog"])} utterances, hit id: {info_dict["hit_id"]}, model_nickname: {model_nickname}.'
                )

            if self.use_problem_buckets:
                dialog_has_problems = False
            for utterance_idx, utt in enumerate(data['dialog']):

                d = {
                    'worker_id': info_dict['worker'],
                    'hit_id': info_dict['hit_id'],
                    'model_nickname': model_nickname,
                    'conversation_idx': conversation_idx,
                    'turn_idx': utterance_idx,
                    'agent_idx': utt['agent_idx'],
                    'text': utt['text'],
                    **{bucket: '' for bucket in self.problem_buckets},
                }

                if utt['agent_idx'] == 1:

                    d['final_rating'] = utt.get('final_rating')

                    if self.use_problem_buckets:
                        if 'problem_data' not in utt:
                            for bucket in self.problem_buckets:
                                d[bucket] = 'MALFORMED'
                            print(
                                f'Warning got MALFORMED utterance problem data inside complete convo: {utt}. Skipping.'
                            )
                            continue
                        else:
                            for bucket in self.regular_buckets + ['none_all_good']:
                                d[bucket] = utt['problem_data'][bucket]
                        for k in self.regular_buckets + ['none_all_good']:
                            if k not in stat_counts[model_nickname]:
                                stat_counts[model_nickname][k] = 0
                            stat_counts[model_nickname][k] += d[k]
                            if k != 'none_all_good' and d[k]:
                                dialog_has_problems = True

                    if 'total' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['total'] = 0
                    if d['agent_idx'] == 1:
                        stat_counts[model_nickname]['total'] += 1
                    if d['final_rating'] is not None:
                        # Only one the last utterance (agent idx == 1)
                        if 'count_ratings' not in stat_counts[model_nickname]:
                            stat_counts[model_nickname]['count_ratings'] = 0
                        stat_counts[model_nickname]['count_ratings'] += 1
                        if 'ratings' not in stat_counts[model_nickname]:
                            stat_counts[model_nickname]['ratings'] = []
                        stat_counts[model_nickname]['ratings'].append(
                            int(d['final_rating'])
                        )

                else:

                    # Counting some aspects of the human's utterances
                    if 'human_utterance_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_utterance_count'] = 0
                    stat_counts[model_nickname]['human_utterance_count'] += 1

                    if 'human_word_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_word_count'] = 0
                    stat_counts[model_nickname]['human_word_count'] += len(
                        d['text'].strip().split(' ')
                    )

                    if 'human_question_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_question_count'] = 0
                    stat_counts[model_nickname]['human_question_count'] += d[
                        'text'
                    ].count('?')

                d = self._add_additional_per_turn_stats(d=d, utt=utt)

                df = df.append(d, ignore_index=True)

            if info_dict['worker'] not in worker_stats:
                worker_stats[info_dict['worker']] = {'conversations': 0}
                if self.use_problem_buckets:
                    worker_stats[info_dict['worker']]['problems_found'] = 0
            worker_stats[info_dict['worker']]['conversations'] += 1

            if self.use_problem_buckets:
                # Count the number of problems the worker got
                is_problem = ~df['none_all_good'].replace('', True)
                # Only want to count bot utterances but human ones, while included,
                # won't be False
                count = is_problem.sum()
                worker_stats[info_dict['worker']]['problems_found'] += count

            # Logic for calculating percent of conversations that are clean
            if 'count_convos' not in stat_counts[model_nickname]:
                stat_counts[model_nickname]['count_convos'] = 0
            stat_counts[model_nickname]['count_convos'] += 1

            if self.use_problem_buckets and not dialog_has_problems:
                if 'convo_clean' not in stat_counts[model_nickname]:
                    stat_counts[model_nickname]['convo_clean'] = 0
                stat_counts[model_nickname]['convo_clean'] += 1

            # Adding the full conversation to the list of conversations
            conversation_dfs.append(df)
            conversation_idx += 1

        for m, conversation_count in complete_convos_per_model.items():
            print(f'Got {conversation_count} complete conversation(s) for model: {m}')

        print(f'{num_complete_convos:d} complete conversation(s) collected.')
        print(f'{len(bad_conversations):d} bad conversation(s).')
        num_approved_convos = num_complete_convos - len(bad_conversations)
        print(f'{num_approved_convos:d} approved conversation(s).')
        print(f'({num_wrong_status_convos:d} wrong status conversation(s) collected.)')
        print(
            f'({num_convos_with_no_save_data:d} conversation(s) collected with no saved data.)'
        )
        for model_nickname, model_stats_dict in stat_counts.items():
            print(f'---{model_nickname}---')
            for p, v in model_stats_dict.items():
                if p == 'count_ratings':
                    continue
                if p == 'ratings':
                    print(
                        f'Average Engaging-ness Rating: {np.average(model_stats_dict["ratings"])} ({model_stats_dict["count_ratings"]} ratings)'
                    )
                    continue
                if p == 'human_word_count' or p == 'human_question_count':
                    print(
                        f'{p}: {v} ({v/model_stats_dict["human_utterance_count"]:.3})'
                    )
                elif p == 'human_utterance_count':
                    print(f'{p}: {v}')
                elif p == 'count_convos':
                    print(f'{p}: {v}')
                elif self.use_problem_buckets and p == 'convo_clean':
                    print(f'{p}: {v} ({v/model_stats_dict["count_convos"]:.2%})')
                else:
                    print(f'{p}: {v} ({v/model_stats_dict["total"]:.2%})')

        print('Printing worker IDs not already in block list to add...')
        for b in bad_conversations:
            worker_id = b['workers'][0]
            if worker_id not in self.worker_block_list:
                print(f"""'{worker_id}',""")
        print('Done printing bad workers.')

        print('Worker stats:')
        worker_columns = ['worker_id', 'conversations']
        if self.use_problem_buckets:
            worker_columns += ['problems_found', 'avg_problems_per_convo']
        worker_df = pd.DataFrame([], columns=worker_columns)

        for worker_id, data in worker_stats.items():
            print(worker_id)

            stat = {'worker_id': worker_id, 'conversations': data['conversations']}
            if self.use_problem_buckets:
                avg_problems_per_convo = data['problems_found'] / data['conversations']
                stat.update(
                    {
                        'problems_found': data['problems_found'],
                        'avg_problems_per_convo': avg_problems_per_convo,
                    }
                )
            worker_df = worker_df.append(stat, ignore_index=True)
        if self.use_problem_buckets:
            worker_df = worker_df.sort_values('avg_problems_per_convo', ascending=0)
        worker_df.to_csv(worker_results_file, index=False)
        print(worker_df)
        print(f'Wrote worker statistical results to: {worker_results_file}')

        # Save full results
        all_conversations_df = pd.DataFrame()
        for df in conversation_dfs:
            all_conversations_df = all_conversations_df.append(df)
        print(f'\nWorker conversation counts: {worker_conversation_counts}')

        return all_conversations_df

    def _add_additional_per_turn_stats(self, d: dict, utt: dict) -> dict:
        """
        Add in additional statistics on the level of each conversation turn.

        Useful for subclasses.
        """
        _ = utt  # utt is ignored in this passthrough method
        return d


if __name__ == '__main__':
    parser_ = ModelChatResultsCompiler.setup_args()
    args_ = parser_.parse_args()
    ModelChatResultsCompiler(vars(args_)).compile_and_save_results()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import yaml
from dataclasses import dataclass, field
from typing import Dict, TYPE_CHECKING

from mephisto.abstractions.blueprint import SharedTaskState
from mephisto.operations.registry import register_mephisto_abstraction
from omegaconf import DictConfig

from parlai.crowdsourcing.tasks.model_chat.model_chat_blueprint import (
    ModelChatBlueprint,
    ModelChatBlueprintArgs,
    BaseModelChatBlueprint,
)
from parlai.crowdsourcing.tasks.pairwise_per_turn_eval.bot_agent import (
    PerTurnEvalTurkLikeAgent,
)

if TYPE_CHECKING:
    from mephisto.data_model.task import TaskRun


def get_task_path():
    return os.path.dirname(os.path.realpath(__file__))


BLUEPRINT_TYPE = 'per_turn_eval_blueprint'


@dataclass
class PerTurnEvalBlueprintArgs(ModelChatBlueprintArgs):
    _blueprint_type: str = BLUEPRINT_TYPE

    custom_source_dir: str = field(
        default=os.path.join(get_task_path(), 'frontend'),
        metadata={"help": "Path to frontend code"},
    )

    random_seed: int = field(
        default=42, metadata={"help": 'Seed for random operations'}
    )

    task_question: str = field(
        default="Which next response from your partner would be more interesting?",
        metadata={"help": "The task question to choose bot's responses based on"},
    )

    task_config_path: str = field(
        default=os.path.join(get_task_path(), 'task_config'),
        metadata={"help": "Base path to pull task configuration information"},
    )

    task_description_file: str = field(
        default="${mephisto.blueprint.task_config_path}/task_description.html",
        metadata={"help": "Path to file of HTML to show on the task-description page"},
    )

    left_pane_text_path: str = field(
        default="${mephisto.blueprint.task_config_path}/left_pane_text.html",
        metadata={
            "help": "Path to file of HTML to show on the left-hand pane of the chat window"
        },
    )


@register_mephisto_abstraction()
class PerTurnEvalBlueprint(ModelChatBlueprint):
    """
    Blueprint for per turn evaluation method.

    This blueprint subclasses ModelChatBlueprint to provide logic for keeping track of
    how many more conversations are needed per model; this logic is not shared with
    other blueprints.
    """

    ArgsClass = PerTurnEvalBlueprintArgs
    BLUEPRINT_TYPE = BLUEPRINT_TYPE

    @classmethod
    def assert_task_args(
        cls, args: "DictConfig", shared_state: "SharedTaskState"
    ) -> None:
        """
        Ensure that arguments are properly configured to launch this task.
        """

        if (
            not isinstance(shared_state.conversations_needed, dict)
            or len(shared_state.conversations_needed) == 0
        ):
            assert (
                args.blueprint.get('conversations_needed_string', None) is not None
            ), (
                "Must provide a string of needed conversations per model if not providing "
                "a conversations needed dict"
            )
            try:
                conversations_needed = {}
                parts = args.blueprint.conversations_needed_string.split(',')
                for part in parts:
                    model1, model2, num_string = part.split(':')
                    models_alphabetize = [model1, model2]
                    models_alphabetize.sort()
                    conversations_needed[
                        f'{models_alphabetize[0]}:{models_alphabetize[1]}'
                    ] = int(num_string)
            except Exception as e:
                raise Exception(
                    "Could not create conversations needed dict from given string. "
                    f"Error was {e}.\n"
                    "Be sure the format is like modelA:modelB:50,modelC:modelD:20"
                )
        else:
            conversations_needed = shared_state.conversations_needed
        args.blueprint.num_conversations = sum(conversations_needed.values())
        BaseModelChatBlueprint.assert_task_args(args=args, shared_state=shared_state)

        if args.blueprint.get("annotations_config_path", "") != "":
            # We are going to do annotations, so check for the presence of an onboarding
            # data file that will be used to onboard users into knowing how to do the
            # annotations properly
            assert (
                args.blueprint.get("onboard_task_data_path", None) is not None
            ), "Must provide an onboarding data file"
            full_path = os.path.expanduser(args.blueprint.onboard_task_data_path)
            assert os.path.exists(
                full_path
            ), f"Target onboarding data path {full_path} doesn't exist"

    def __init__(
        self, task_run: "TaskRun", args: "DictConfig", shared_state: "SharedTaskState"
    ):
        conversations_needed = self._process_conversations_needed(args)
        self.conversations_needed = conversations_needed
        shared_state.conversations_needed = conversations_needed
        args.blueprint.num_conversations = sum(conversations_needed.values())

        shared_state.world_opt.update({'task_question': args.blueprint.task_question})

        super().__init__(task_run=task_run, args=args, shared_state=shared_state)

    def format_left_pane_text(self, args: "DictConfig"):
        """
        Adds the user's intended persona into the left pane of the frontend by modifying
        self.left_pane_text for code injection.
        """
        self.left_pane_text = self.left_pane_text.format(
            task_question=args.blueprint.task_question
        )

    def _process_conversations_needed(self, args: "DictConfig") -> Dict[str, int]:
        """
        Formats the pair of models and sets the number of conversations needed.
        """

        conversations_needed_string = args.blueprint.conversations_needed_string
        conversations_needed = {}
        parts = conversations_needed_string.split(',')
        for part in parts:
            model1, model2, num_string = part.split(':')
            models_alphabetize = [model1, model2]
            models_alphabetize.sort()  # alphabetizing for consistency
            conversations_needed[
                f'{models_alphabetize[0]}:{models_alphabetize[1]}'
            ] = int(
                num_string
            )  # format is model_1_name:model_2_name

        return conversations_needed

    def _get_shared_models(self, args: "DictConfig") -> Dict[str, dict]:
        """
        Loads the appropriate model pair and sends it into bot agent.
        """

        with open(args.blueprint.model_opt_path) as f:
            all_model_opts = yaml.safe_load(f.read())

        active_model_opts = {}
        for model_pair in self.conversations_needed:
            for model, opt in all_model_opts.items():
                if (
                    model in model_pair.split(':')
                    and self.conversations_needed[model_pair] > 0
                ):
                    active_model_opts[model] = opt

        return PerTurnEvalTurkLikeAgent.get_bot_agents(
            args=args, model_opts=active_model_opts
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'https://raw.githubusercontent.com/kushalchawla/CaSiNo/main/data/split/casino_train.json',
        'casino_train.json',
        '6b953d153fc8c78f27e911c1439b93b9b3519357e3ba825091b2e567845ba3a7',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/kushalchawla/CaSiNo/main/data/split/casino_valid.json',
        'casino_valid.json',
        '91f2d1f09accedf98667ac081fd5083752738390734e991601b036643da077e0',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/kushalchawla/CaSiNo/main/data/split/casino_test.json',
        'casino_test.json',
        'bf6da2d7c105396300d85a65819c04d99304ac9abb8a590ba342fd0c86b4dd12',
        zipped=False,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'casino')
    version = "v1.1"

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark as done
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
File that includes tests for teacher metrics.
"""

import unittest
from math import isnan

from parlai.core.metrics import AverageMetric
from parlai.core.tod.teacher_metrics import SlotF1Metric, SlotMetrics


class TestSlotF1Metric(unittest.TestCase):
    """
    Test SlotF1Metric.
    """

    def test_slot_f1_metric_inputs(self):
        slots_p_r_and_f1 = [
            (None, None, float("nan")),
            (None, AverageMetric(0.0), float("nan")),
            (AverageMetric(0.0), AverageMetric(0.0), float("nan")),
            (AverageMetric(1), AverageMetric(1), 1.0),
            (AverageMetric(1), AverageMetric(0), 0.0),
            (AverageMetric(0.25), AverageMetric(0.75), 0.375),
        ]
        for slot_p, slot_r, slot_f1 in slots_p_r_and_f1:
            actual_slot_f1 = SlotF1Metric(slot_p=slot_p, slot_r=slot_r).value()
            if isnan(slot_f1):
                self.assertTrue(isnan(actual_slot_f1))
            else:
                self.assertEqual(slot_f1, actual_slot_f1)

    def test_slot_f1_metric_addition(self):
        a = SlotF1Metric(slot_p=1)
        b = SlotF1Metric(slot_r=0)
        c = SlotF1Metric(slot_p=AverageMetric(numer=2, denom=3), slot_r=1)
        d = a + b + c
        # Slot P should be 3/4 = 0.75; slot R should be 1/2 = 0.5
        self.assertEqual(0.6, d.value())


empty_slots = {}
basic_slots = {"a": "a_val", "b": "b_val", "c": "c_val"}
partial_slots = {"a": "a_val", "other": "other_val"}


class TestSlotMetrics(unittest.TestCase):
    def test_base_slot_metrics(self):
        cases = [
            (empty_slots, empty_slots, {"jga": 1}),
            (
                basic_slots,
                basic_slots,
                {"jga": 1, "slot_p": 1, "slot_r": 1, "slot_f1": 1},
            ),
            (
                basic_slots,
                partial_slots,
                {"jga": 0, "slot_p": 0.5, "slot_r": float(1.0 / 3), "slot_f1": 0.4},
            ),
        ]
        for teacher, predicted, result in cases:
            metric = SlotMetrics(teacher_slots=teacher, predicted_slots=predicted)
            for key in result:
                self.assertEqual(result[key], metric.report()[key])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict, List, Union, TYPE_CHECKING

from mephisto.operations.registry import register_mephisto_abstraction
from mephisto.abstractions.blueprint import SharedTaskState
from mephisto.abstractions.blueprints.static_react_task.static_react_blueprint import (
    StaticReactBlueprint,
    StaticReactBlueprintArgs,
)
from omegaconf import DictConfig

if TYPE_CHECKING:
    from mephisto.data_model.task import TaskRun


def get_task_path():
    return os.path.dirname(__file__)


STATIC_BLUEPRINT_TYPE = 'dialcrowd_static_blueprint'


@dataclass
class DialCrowdStaticBlueprintArgs(StaticReactBlueprintArgs):
    _blueprint_type: str = STATIC_BLUEPRINT_TYPE
    _group: str = field(
        default="DialCrowdStaticBlueprint",
        metadata={
            'help': """This task renders conversations from a file and asks for turn by turn annotations of them."""
        },
    )
    subtasks_per_unit: int = field(
        default=-1, metadata={"help": "Number of subtasks/comparisons to do per unit"}
    )


@register_mephisto_abstraction()
class DialCrowdStaticBlueprint(StaticReactBlueprint):
    """
    This Blueprint has a subtasks number option to combine multiple conversations into
    "sub-HITs".

    It also has options for the onboarding data answers and the annotation bucket
    definitions.
    """

    _initialization_data_dicts: Union[List[List[Dict[str, Any]]], List[Dict[str, Any]]]

    ArgsClass = DialCrowdStaticBlueprintArgs
    BLUEPRINT_TYPE = STATIC_BLUEPRINT_TYPE

    def __init__(
        self, task_run: "TaskRun", args: "DictConfig", shared_state: "SharedTaskState"
    ):
        super().__init__(task_run, args=args, shared_state=shared_state)
        self.subtasks_per_unit = self.args.blueprint.subtasks_per_unit

        if self.subtasks_per_unit <= 0:
            raise Exception(
                f'subtasks_per_unit must be greater than zero but was {self.subtasks_per_unit}'
            )

        self.raw_data = self._initialization_data_dicts

        # Now chunk the data into groups of <num_subtasks>
        grouped_data = []
        logging.info(
            f'Raw data length: {len(self.raw_data)}. self.subtasks_per_unit: {self.subtasks_per_unit}'
        )
        for i in range(0, len(self._initialization_data_dicts), self.subtasks_per_unit):
            chunk = self._initialization_data_dicts[i : i + self.subtasks_per_unit]
            grouped_data.append(chunk)
        self._initialization_data_dicts = grouped_data  # type: ignore
        # Last group may have less unless an exact multiple
        logging.info(
            f'Grouped data into {len(self._initialization_data_dicts)} tasks with {self.subtasks_per_unit} subtasks each.'
        )

    def get_frontend_args(self) -> Dict[str, Any]:
        """
        Specifies what options within a task_config should be forwarded to the client
        for use by the task's frontend.
        """

        # load the task configuration
        with open(os.path.join(get_task_path(), 'task_config/config.json')) as f:
            task_config = json.load(f)

        # combine the task configuration loaded from json with the settings
        # required by ParlAI.
        task_config.update(
            {
                "task_description": self.args.task.get('task_description', None),
                "task_title": self.args.task.get('task_title', None),
                "frame_height": 0,
                "num_subtasks": self.args.blueprint.subtasks_per_unit,
                "block_mobile": True,
            }
        )
        return task_config



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Helpers so we don't need to create agents all over.
"""

import parlai.core.tod.tod_agents as tod_agents
import parlai.core.tod.tod_core as tod_core

import os

API_DATABASE_FILE = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "standalone_api_file.pickle"
)


def episode_has_broken_api_turn(episode_idx, max_turns):
    """
    To generate predictably broken episodes (for calculating metrics)
    """
    return episode_idx % 2 == 1 and max_turns > 0


def turn_has_broken_api_call(round_idx, episode_idx):
    """
    To generate predictably broken turns (for calculating metrics)
    """
    return episode_idx % 2 == 1 and round_idx % 3 == 1


def make_api_call_machine(round_idx, episode_idx=0, use_broken_mock_api_calls=False):
    if round_idx == 0:
        return {}
    if use_broken_mock_api_calls:
        # Hack as a way to test metrics reporting in tod world script
        if turn_has_broken_api_call(round_idx, episode_idx):
            round_idx = -1 * round_idx
    return {tod_core.STANDARD_API_NAME_SLOT: f"name_{round_idx}", "in": round_idx}


def make_api_resp_machine(round_idx):
    if round_idx == 0:
        return {}
    return {"out": round_idx}


def make_api_schemas_machine(max_rounds):
    return [
        {
            tod_core.STANDARD_API_NAME_SLOT: f"name_{round_idx}",
            tod_core.STANDARD_REQUIRED_KEY: ["in"],
            tod_core.STANDARD_OPTIONAL_KEY: [],
        }
        for round_idx in range(1, max_rounds)
    ]


def make_goal_calls_machine(max_rounds):
    return [make_api_call_machine(x) for x in range(1, max_rounds)]


def get_rounds(episode_idx, max_rounds, use_broken_mock_api_calls=False):
    return [
        tod_core.TodStructuredRound(
            user_utt=f"user_utt_{episode_idx}_{round_idx}",
            api_call_machine=make_api_call_machine(
                round_idx, episode_idx, use_broken_mock_api_calls
            ),
            api_resp_machine=make_api_resp_machine(round_idx),
            sys_utt=f"sys_utt_{episode_idx}_{round_idx}",
        )
        for round_idx in range(max_rounds)
    ]


def get_round_utts(episode_idx, max_rounds, filter_utts=None):
    if max_rounds < 1:
        return []
    utts = [
        [
            f"{tod_core.STANDARD_USER_UTTERANCE}user_utt_{episode_idx}_0",
            tod_core.STANDARD_CALL,
            tod_core.STANDARD_RESP,
            f"{tod_core.STANDARD_SYSTEM_UTTERANCE}sys_utt_{episode_idx}_0",
        ]
    ]
    for i in range(1, max_rounds):
        utts.append(
            [
                f"{tod_core.STANDARD_USER_UTTERANCE}user_utt_{episode_idx}_{i}",
                f"{tod_core.STANDARD_CALL}api_name = name_{i} ; in = {i}",
                f"{tod_core.STANDARD_RESP}out = {i}",
                f"{tod_core.STANDARD_SYSTEM_UTTERANCE}sys_utt_{episode_idx}_{i}",
            ]
        )
    utts.append(
        [
            f"{tod_core.STANDARD_USER_UTTERANCE}{tod_core.STANDARD_DONE}",
            tod_core.STANDARD_CALL,
            tod_core.STANDARD_RESP,
            tod_core.STANDARD_SYSTEM_UTTERANCE,
        ]
    )
    if filter_utts is not None:
        utts = [
            [turn for i, turn in enumerate(round_data) if filter_utts[i]]
            for round_data in utts
        ]
    return utts


TEST_NUM_EPISODES_OPT_KEY = "test_num_episodes"
TEST_NUM_ROUNDS_OPT_KEY = "test_num_rounds"

# No api calls in this setup
EPISODE_SETUP__UTTERANCES_ONLY = {
    TEST_NUM_ROUNDS_OPT_KEY: 1,
    TEST_NUM_EPISODES_OPT_KEY: 1,
}

# Only one call, one goal, one api description in this setup
EPISODE_SETUP__SINGLE_API_CALL = {
    TEST_NUM_ROUNDS_OPT_KEY: 2,
    TEST_NUM_EPISODES_OPT_KEY: 1,
}
# Will start testing multiple api calls + schemas, multi-round logic
EPISODE_SETUP__MULTI_ROUND = {TEST_NUM_ROUNDS_OPT_KEY: 5, TEST_NUM_EPISODES_OPT_KEY: 1}

# Test that episode logic is correct
EPISODE_SETUP__MULTI_EPISODE = {
    TEST_NUM_ROUNDS_OPT_KEY: 5,
    TEST_NUM_EPISODES_OPT_KEY: 8,
}

# Test that episode + pesky-off-by-one batchinglogic is correct
EPISODE_SETUP__MULTI_EPISODE_BS = {
    TEST_NUM_ROUNDS_OPT_KEY: 5,
    TEST_NUM_EPISODES_OPT_KEY: 35,
}


class TestDataParser(tod_agents.TodStructuredDataParser):
    """
    Assume that when we init, we init w/ num of episodes + rounds as opts.
    """

    def __init__(self, opt, shared=None):
        opt["datafile"] = "DUMMY"
        self.fold = "DUMMY"
        # Following lines are only relevant in training the standalone api teacher
        if TEST_NUM_EPISODES_OPT_KEY not in opt:
            opt[TEST_NUM_EPISODES_OPT_KEY] = 35
        if TEST_NUM_ROUNDS_OPT_KEY not in opt:
            opt[TEST_NUM_ROUNDS_OPT_KEY] = 5
        super().__init__(opt, shared)

    def setup_episodes(self, _):
        result = []
        for ep_idx in range(0, self.opt[TEST_NUM_EPISODES_OPT_KEY]):
            result.append(
                tod_core.TodStructuredEpisode(
                    goal_calls_machine=[
                        make_api_call_machine(x)
                        for x in range(1, self.opt[TEST_NUM_ROUNDS_OPT_KEY])
                    ],
                    api_schemas_machine=make_api_schemas_machine(
                        self.opt[TEST_NUM_ROUNDS_OPT_KEY]
                    ),
                    rounds=get_rounds(
                        ep_idx,
                        self.opt[TEST_NUM_ROUNDS_OPT_KEY],
                        self.opt.get("use_broken_mock_api_calls", False),
                    ),
                )
            )
        return result

    def get_id_task_prefix(self):
        return "Test"


class SystemTeacher(TestDataParser, tod_agents.TodSystemTeacher):
    pass


class UserSimulatorTeacher(TestDataParser, tod_agents.TodUserSimulatorTeacher):
    pass


class StandaloneApiTeacher(TestDataParser, tod_agents.TodStandaloneApiTeacher):
    pass


class GoalAgent(TestDataParser, tod_agents.TodGoalAgent):
    pass


class ApiSchemaAgent(TestDataParser, tod_agents.TodApiSchemaAgent):
    pass


class SingleGoalAgent(TestDataParser, tod_agents.TodSingleGoalAgent):
    pass


class SingleApiSchemaAgent(TestDataParser, tod_agents.TodSingleApiSchemaAgent):
    pass


# Tested in tod world code
class UserUttAgent(TestDataParser, tod_agents.TodUserUttAgent):
    pass


# Tested in tod world code
class ApiCallAndSysUttAgent(TestDataParser, tod_agents.TodApiCallAndSysUttAgent):
    pass



from parlai.core.teachers import DialogTeacher
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt

from typing import Optional

from .build import build
import random
import json
import os


class SPOLINDialogueTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        self.id = "SPOLIN"
        self.datatype = opt['datatype']
        build(opt)
        suffix = 'train' if opt['datatype'].startswith('train') else 'valid'
        if opt.get('use_acl_version') and suffix == 'train':
            suffix += '-acl'

        opt['datafile'] = os.path.join(
            opt['datapath'], 'spolin', f"spolin-{suffix}.json"
        )
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('SPOLIN Dialogue Arguments')
        agent.add_argument(
            '-acl',
            '--use_acl_version',
            action="store_true",
            help='Use the version in the ACL paper that does not include further augmentation from SubTle Corpus',
        )

        agent.add_argument(
            '--include_nonyesands',
            action="store_true",
            help='Include non-yesands',
        )

        return parser

    def setup_data(self, path):

        print(f"Loading: {path}")
        with open(path, "r") as f:
            # don't set self.data
            self.data_ = json.load(f)

        processed_data = []
        yesands_dict = self.data_['yesands']
        for _source, yas in yesands_dict.items():
            processed_data += yas

        if self.opt.get("include_nonyesands"):
            non_yesands_dict = self.data['non-yesands']
            for _source, nyas in non_yesands_dict.items():
                processed_data += nyas

        self.processed_data = processed_data
        if 'train' in self.datatype:
            random.shuffle(self.processed_data)

        for ya_pair in self.processed_data:
            new_episode = True
            yield {
                "text": ya_pair['p'],
                "labels": ya_pair['r'],
            }, new_episode


class DefaultTeacher(SPOLINDialogueTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Imported from https://github.com/harsh19/Reasoning-Chains-MultihopQA
Paper: "Learning to Explain: Datasets and Models for Identifying Valid Reasoning Chains in Multihop Question-Answering"
(Jhamtani and Clark 2020) https://aclanthology.org/2020.emnlp-main.10.pdf

Multihop Explanations for QASC dataset.
"""

import json
import os

from parlai.utils.data import DatatypeHelper

from .build import build
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.io import PathManager
from typing import Optional

from tasks.reasoning.agents import StepByStepReasoningTeacher
import random


class EqascStepByStepReasoningTeacher(StepByStepReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following common sence problem. Show your work."
        )
        parser.add_argument(
            "--extrinsic-step",
            type=bool,
            default=False,
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)  # NOTE: the call to build here
        opt['datafile'] = os.path.join(opt['datapath'], 'eqasc')
        self.id = 'eqasc'
        self.fold = DatatypeHelper.fold(opt['datatype'])
        self.eqasc_random = random.Random(42)
        super().__init__(opt, shared)

    def get_data_for_fold(self, fold):
        extrinsic_step = self.opt.get("extrinsic_step", False)
        suffix = self.fold
        if self.fold == "valid":
            suffix = "dev"
        filename = os.path.join(self.opt['datafile'], f"eqasc_{suffix}_grc.json")

        json_data = json.load(PathManager.open(filename))

        if len(json_data) == 0:
            return

        data = []
        for example in json_data:
            question = example['question']['stem'].strip()
            f1 = example['fact1']
            f2 = example['fact2']
            steps = [f1, f2]
            answer = example['combinedfact']
            m = {
                "question": question,
                "answer": answer,
                "steps": steps,
            }
            if not extrinsic_step:
                yield m
            else:
                data.append(m)

        if extrinsic_step:
            for line in data:
                rand_steps = self.eqasc_random.choice(data)["steps"]
                random_step = self.eqasc_random.choice(rand_steps)
                yield {
                    "question": line["question"],
                    "answer": line["answer"],
                    "steps": line["steps"],
                    "extrinsic_step": random_step,
                }


class DefaultTeacher(EqascStepByStepReasoningTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile

import csv
from itertools import islice
from pathlib import Path
import os
import json
import re
import tqdm

DEBUG_MISSING_RAW_CONVERSATIONS = False  # Unnecessary once Amazon fixes multidogo

RESOURCE = DownloadableFile(
    "https://github.com/awslabs/multi-domain-goal-oriented-dialogues-dataset/archive/master.zip",
    "raw_data.zip",
    "fb59c7261da2d30d9d24b9af309ebb4bf0e5b39f97d718201a7160e591e76a3c",
    zipped=True,
)

RAW_DATA_PREFIX = "multi-domain-goal-oriented-dialogues-dataset-master/data/"

RAW_DATA_ANNOTATED_DATA_PATH = "paper_splits"
RAW_DATA_UNANNOTATED_DATA_PATH = "unannotated"

TURN_INTENT = "turn"
SENTENCE_INTENT = "sentence"
TURN_AND_SENTENCE_INTENT = "both"

RAW_DATA_SENTENCE_INTENT_PATH = "splits_annotated_at_sentence_level"
RAW_DATA_TURN_INTENT_PATH = "splits_annotated_at_turn_level"

RAW_DATA_INTENT_BY_TYPE_PATH = {
    TURN_INTENT: RAW_DATA_TURN_INTENT_PATH,
    SENTENCE_INTENT: RAW_DATA_SENTENCE_INTENT_PATH,
}

DOMAINS = ["airline", "fastfood", "finance", "insurance", "media", "software"]

DATATYPE_TO_RAW_DATA_FILE_NAME = {
    "test": "test.tsv",
    "train": "train.tsv",
    "valid": "dev.tsv",
}

PROCESSED = "processed/"


def _preprocess(opt, datapath, datatype, version):
    """
    MultiDoGo conversations take place between an "agent" and a customer". Labeled
    customer data is stored in one set of files while the agent data is in another.
    There is a common conversation ID between the two, but the conversations are not
    listed in a consistent way between the documents. Since we'll have to do work to
    associate the data between the files anyway, we might as well process the data into
    a new file that'll be easier to deal with.

    Stores the data as <multidogo_data_path>/processed/<domain>/<datatype>.txt.
    Will skip preprocessing if this file already exists.
    """
    domains = opt.get("domains", DOMAINS)
    intent_type = opt.get("intent_type", TURN_INTENT)

    for domain in domains:
        out_dir = get_processed_multidogo_folder(
            datapath, domain, datatype, intent_type
        )
        if build_data.built(out_dir, version):
            continue
        print(
            f"    Preprocessing '{domain}' data for '{datatype}' with '{intent_type}' intent labels."
        )

        Path(out_dir).mkdir(parents=True, exist_ok=True)

        # The agent responses for *all* datatypes are in one file.
        # We need to iterate through the datatype file to know which lines
        # we'll actually need... so build a quick lookup table to know which
        # lines in the tsv file we'll need to care about so we're not scanning
        # through the whole thing a bunch
        unannotated_id_map = _build_conversation_span_map(
            _get_unannotated_tsv_data(datapath, domain)
        )

        # Actually do the work of collating all of the conversations + annotations
        # For turn + sentence intent labels, we do two passes, one for sentence
        # then one for turn so that we do not add two sets of labels for the
        # same conversation ID. We can use this forced structure to do the
        # separate categories of turn intent and sentence intent labels.  We
        # also do a bit of chuking
        file_idx = 0
        seen_conversations_set = set()
        if intent_type == TURN_AND_SENTENCE_INTENT or intent_type == SENTENCE_INTENT:
            file_idx, seen_conversations_set = _aggregate_and_write_conversations(
                intent_type,
                SENTENCE_INTENT,
                datapath,
                domain,
                datatype,
                unannotated_id_map,
                start_file_idx=file_idx,
                skip_ids=set(),
            )

        if intent_type == TURN_AND_SENTENCE_INTENT or intent_type == TURN_INTENT:
            _, _ = _aggregate_and_write_conversations(
                intent_type,
                TURN_INTENT,
                datapath,
                domain,
                datatype,
                unannotated_id_map,
                start_file_idx=file_idx,
                skip_ids=seen_conversations_set,
            )

        # mark that we've built this combinations
        build_data.mark_done(out_dir, version_string=version)


def get_processed_multidogo_folder(datapath, domain, datatype, intent_type):
    return os.path.join(datapath, PROCESSED, domain, intent_type, datatype)


# unannotated data is UNANNOTATED_DATA_PROFIX + <domain> + '.tsv'
# annotated data is ANNOTATED_DATA_PATH + <annotations type> + <domain> + '/' + <datatype> + '.tsv'
def _get_unannotated_tsv_data(datapath, domain):
    file_name = os.path.join(
        datapath, RAW_DATA_PREFIX, RAW_DATA_UNANNOTATED_DATA_PATH, domain + ".tsv"
    )
    return csv.reader(open(file_name, "r"), delimiter=",")  # comma-separated tsv, lol


def _get_annotated_tsv_data(datapath, domain, datatype, annotation_type):
    file_name = os.path.join(
        datapath,
        RAW_DATA_PREFIX,
        RAW_DATA_ANNOTATED_DATA_PATH,
        RAW_DATA_INTENT_BY_TYPE_PATH[annotation_type],
        domain,
        DATATYPE_TO_RAW_DATA_FILE_NAME[datatype],
    )
    return csv.reader(open(file_name, "r"), delimiter="\t")


def _get_annotated_tsv_data_size(datapath, domain, datatype, annotation_type):
    file_name = os.path.join(
        datapath,
        RAW_DATA_PREFIX,
        RAW_DATA_ANNOTATED_DATA_PATH,
        RAW_DATA_INTENT_BY_TYPE_PATH[annotation_type],
        domain,
        DATATYPE_TO_RAW_DATA_FILE_NAME[datatype],
    )
    return sum(1 for line in open(file_name, 'r'))


def _build_conversation_span_map(unannotated_tsv_object):
    result = {}  # conversationId to (start line, length) map
    start = 0
    prev_conversation_id = ""
    length = 0
    for i, row in enumerate(unannotated_tsv_object):
        conversation_id = row[0][
            4:-2
        ]  # do substring cause conversationId has extra filler in unannotated
        if conversation_id != prev_conversation_id:
            result[prev_conversation_id] = (start, length)
            start = i
            prev_conversation_id = conversation_id
            length = 0
        length += 1
    result[conversation_id] = (start, length)
    return result


def _get_slots_map(utterance, slot_string):
    values = slot_string.split(" ")
    cleaned = re.sub(r"[^\w\s]", "", utterance)
    words = cleaned.split(" ")
    result = {}
    for i in range(len(words)):
        if values[i] != "O":
            result[values[i]] = words[i]
    return result


def _aggregate_and_write_conversations(
    raw_intent_type,
    fetch_intent_type,
    datapath,
    domain,
    datatype,
    unannotated_id_map,
    skip_ids,
    start_file_idx=0,
):
    conversations_to_write = {}  # conversationId -> list of turns
    seen_conversations = set()
    out_dir = get_processed_multidogo_folder(
        datapath, domain, datatype, raw_intent_type
    )
    file_idx = start_file_idx
    intent_tsv = _get_annotated_tsv_data(datapath, domain, datatype, fetch_intent_type)
    next(intent_tsv)  # don't need the header in the first line
    print(f"Processing for {domain}, {fetch_intent_type}, {datatype}")
    for labeled_line in tqdm.tqdm(
        intent_tsv,
        total=_get_annotated_tsv_data_size(
            datapath, domain, datatype, fetch_intent_type
        )
        - 1,
    ):
        conversation_id = labeled_line[0]
        if conversation_id in skip_ids:
            continue
        if conversation_id not in seen_conversations:
            # new conversation, add text of conversation to conversations_to_write
            conversations_to_write[conversation_id] = {}
            found_raw_conversation = _add_utterances(
                unannotated_id_map,
                conversation_id,
                conversations_to_write,
                datapath,
                domain,
            )
            seen_conversations.add(conversation_id)
            if not found_raw_conversation:
                if DEBUG_MISSING_RAW_CONVERSATIONS:
                    print(f"Could not find raw conversations for {conversation_id}")
                skip_ids.add(conversation_id)
                conversations_to_write.pop(conversation_id, None)
                continue
        if fetch_intent_type == SENTENCE_INTENT:
            _get_sentence_labels_and_slots_map(labeled_line, conversations_to_write)
        elif fetch_intent_type == TURN_INTENT:
            _get_turn_labels_and_slots_map(labeled_line, conversations_to_write)
        else:
            raise KeyError(
                "Invalid `fetch_intent_type`. This case should never be hit. Something is broken in the `build.py` file."
            )
    # Don't forget to dump out last file
    with open(f"{out_dir}/{file_idx}.json", "w+") as out_file:
        json.dump(conversations_to_write, out_file, indent=4)
        file_idx += 1
    # Return necessary outputs for next pass
    return file_idx, seen_conversations


def _add_utterances(
    unannotated_id_map, conversation_id, conversations_to_write, datapath, domain
):
    try:
        start, length = unannotated_id_map[conversation_id]
    except KeyError:
        return False
    conversation_text = islice(
        _get_unannotated_tsv_data(datapath, domain), start, start + length
    )

    for line in conversation_text:
        # Format of unannotated: conversationId,turnNumber,utteranceId,utterance,authorRole
        conversations_to_write[conversation_id] = {
            **conversations_to_write[conversation_id],
            int(line[1]): {"text": line[3], "role": line[4]},
        }
    return True


def _get_sentence_labels_and_slots_map(labeled_line, output):
    # Sentence tsv format: conversationId   turnNumber  sentenceNumber  utteranceId utterance   slot-labels intent
    conversation_id = labeled_line[0]
    turn_number = int(float(labeled_line[1]))  # cause a few got saved as float.
    if conversation_id not in output:
        raise RuntimeError("Should never happen; raw conversation text should be here")
    if turn_number not in output[conversation_id]:
        output[conversation_id][turn_number] = {}
    output[conversation_id][turn_number] = {
        **output[conversation_id][turn_number],
        "slots": _get_slots_map(labeled_line[4], labeled_line[5]),
    }
    if "intents" not in output[conversation_id][turn_number]:
        output[conversation_id][turn_number]["intents"] = []
    output[conversation_id][turn_number]["intents"].append(labeled_line[6])


def _get_turn_labels_and_slots_map(labeled_line, output):
    # Turn tsv format: conversationId  turnNumber  utteranceId utterance   slot-labels intent
    conversation_id = labeled_line[0]
    turn_number = int(float(labeled_line[1]))  # cause a few got saved as float
    if conversation_id not in output:
        raise RuntimeError("Should never happen; raw conversation text should be here")
    if turn_number not in output[conversation_id]:
        output[conversation_id][turn_number] = {}
    output[conversation_id][turn_number] = {
        **output[conversation_id][turn_number],
        "slots": _get_slots_map(labeled_line[3], labeled_line[4]),
        "intents": [labeled_line[5]],
    }


def build(opt):
    # get path to data directory
    datapath = os.path.join(opt["datapath"], "multidogo")
    # define version if any
    version = "v1.1"

    # check if data had been previously downloaded
    if not build_data.built(datapath, version_string=version):
        print("[building data: " + datapath + "]")

        # make a clean directory if needed
        if build_data.built(datapath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(datapath)
        build_data.make_dir(datapath)

        # Download the data.
        RESOURCE.download_file(datapath)

        # mark the data as built
        build_data.mark_done(datapath, version_string=version)

    # do preprocessing on the data to put it into FBDialogueData format. There's a lot so check to make sure it's okay
    for fold in ["train", "valid", "test"]:
        _preprocess(opt, datapath, fold, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Utilities for running tests.
"""

import os
import random
import tempfile
import time
import unittest
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import torch
from parlai.utils import logging
from hydra.experimental import compose, initialize
from mephisto.abstractions.blueprint import SharedTaskState
from mephisto.abstractions.databases.local_database import LocalMephistoDB
from mephisto.data_model.task_run import TaskRunArgs
from mephisto.operations.operator import Operator
from mephisto.tools.scripts import augment_config_from_db
from mephisto.utils.metrics import InaccessiblePrometheusServer
from pytest_regressions.data_regression import DataRegressionFixture


class AbstractCrowdsourcingTest:
    """
    Abstract class for end-to-end tests of Mephisto-based crowdsourcing tasks.

    Allows for setup and teardown of the operator, as well as for config specification
    and agent registration.
    """

    def _setup(self):
        """
        To be run before a test.

        Should be called in a pytest setup/teardown fixture.
        """

        random.seed(0)
        np.random.seed(0)
        torch.manual_seed(0)

        self.operator = None
        self.server = None

    def _teardown(self):
        """
        To be run after a test.

        Should be called in a pytest setup/teardown fixture.
        """

        try:
            if self.operator is not None:
                self.operator.force_shutdown()
        except InaccessiblePrometheusServer as e:
            logging.warning(f'InaccessiblePrometheusServer: {e}')

        if self.server is not None:
            self.server.shutdown_mock()

    def _set_up_config(
        self,
        task_directory: str,
        overrides: Optional[List[str]] = None,
        config_name: str = "example",
    ):
        """
        Set up the config and database.

        Uses the Hydra compose() API for unit testing and a temporary directory to store
        the test database.
        :param blueprint_type: string uniquely specifying Blueprint class
        :param task_directory: directory containing the `conf/` configuration folder.
          Will be injected as `${task_dir}` in YAML files.
        :param overrides: additional config overrides
        """

        # Define the configuration settings
        relative_task_directory = os.path.relpath(
            task_directory, os.path.dirname(__file__)
        )
        relative_config_path = os.path.join(
            relative_task_directory, 'hydra_configs', 'conf'
        )
        if overrides is None:
            overrides = []
        with initialize(config_path=relative_config_path):
            self.config = compose(
                config_name=config_name,
                overrides=[
                    f'mephisto/architect=mock',
                    f'mephisto/provider=mock',
                    f'+task_dir={task_directory}',
                    f'+current_time={int(time.time())}',
                ]
                + overrides,
            )

        self.data_dir = tempfile.mkdtemp()
        self.database_path = os.path.join(self.data_dir, "mephisto.db")
        self.db = LocalMephistoDB(self.database_path)
        self.config = augment_config_from_db(self.config, self.db)
        self.config.mephisto.architect.should_run_server = True
        self.config.mephisto.task = TaskRunArgs(
            **self.config.mephisto.task, no_submission_patience=1
        )

    def _set_up_server(self, shared_state: Optional[SharedTaskState] = None):
        """
        Set up the operator and server.
        """
        self.operator = Operator(self.db)
        self.operator.launch_task_run(self.config.mephisto, shared_state=shared_state)
        self.server = self._get_server()

    def _get_live_run(self):
        """
        Get the LiveTaskRun from this job.
        """
        live_runs = list(self.operator.get_running_task_runs().values())
        if len(live_runs) == 0:
            raise ValueError('No live runs present')
        return live_runs[0]

    def _get_server(self):
        """
        Return the MockArchitect's server associated with this run.
        """
        live_run = self._get_live_run()
        return live_run.architect.server

    def _get_channel_info(self):
        """
        Return channel info for the currently running job.
        """
        live_run = self._get_live_run()
        channels = list(live_run.client_io.channels.values())
        if len(channels) > 0:
            return channels[0]
        else:
            raise ValueError('No channel could be detected!')

    def _register_mock_agents(
        self, num_agents: int = 1, assume_onboarding: bool = False
    ) -> List[str]:
        """
        Register mock agents for testing and onboard them if needed, taking the place of
        crowdsourcing workers.

        Specify the number of agents to register. Return the agents' IDs after creation.
        """

        for idx in range(num_agents):

            mock_worker_registration_name = f"MOCK_WORKER_{idx:d}"
            mock_worker_name = f"{mock_worker_registration_name}_sandbox"

            # Register the agent
            mock_agent_details = f"FAKE_ASSIGNMENT_{idx:d}"
            self.server.register_mock_agent(
                mock_worker_registration_name, mock_agent_details
            )
            self.assert_sandbox_worker_created(mock_worker_name)
            workers = self.db.find_workers(worker_name=mock_worker_name)
            print(
                "Workers:",
                workers,
                self.db.find_workers(),
                [w.worker_name for w in self.db.find_workers()],
            )
            worker_id = workers[0].db_id

            if assume_onboarding:
                # Submit onboarding from the agent
                onboard_agents = self.db.find_onboarding_agents()
                onboard_data = {"onboarding_data": {"success": True}}
                self.server.register_mock_agent_after_onboarding(
                    worker_id, onboard_agents[0].get_agent_id(), onboard_data
                )
                self.await_channel_requests()
            self.await_channel_requests()
            print("Agents:", self.db.find_agents())

            # Make sure the agent can be found
            try:
                _ = self.db.find_agents()[idx]
            except IndexError:
                raise ValueError('The agent could not be registered!')

        # Get all agents' IDs
        agents = self.db.find_agents()
        if len(agents) != num_agents:
            raise ValueError(
                f'The actual number of agents is {len(agents):d} instead of the '
                f'desired {num_agents:d}!'
            )
        agent_ids = [agent.db_id for agent in agents]

        return agent_ids

    def await_channel_requests(self, timeout=2) -> None:
        time.sleep(0.1)
        tracked_run = self._get_live_run()
        assert self.operator._run_loop_until(
            lambda: len(tracked_run.client_io.request_id_to_channel_id) == 0, timeout
        ), f"Channeled requests not processed in time!"

    def assert_sandbox_worker_created(self, worker_name, timeout=2) -> None:
        assert self.operator._run_loop_until(
            lambda: len(self.db.find_workers(worker_name=worker_name)) > 0, timeout
        ), f"Worker {worker_name} not created in time!"


class AbstractOneTurnCrowdsourcingTest(AbstractCrowdsourcingTest):
    """
    Abstract class for end-to-end tests of one-turn crowdsourcing tasks.

    Useful for Blueprints such as AcuteEvalBlueprint and StaticReactBlueprint for which
    all of the worker's responses are sent to the backend code at once.
    """

    def _test_agent_state(
        self, task_data: Dict[str, Any], data_regression: DataRegressionFixture
    ):
        """
        Test that the actual agent state matches the expected state.

        Get the final agent state given the input task data and check that it is as
        expected.
        """
        state = self._get_agent_state(task_data=task_data)
        self._check_agent_state(state=state, data_regression=data_regression)

    def _get_agent_state(self, task_data: Dict[str, Any]):
        """
        Submit user task data and return the final agent state.

        Register a mock human agent, request initial data to define the 'inputs' field
        of the agent state, make the agent act to define the 'outputs' field of the
        agent state, and return the agent state.
        """

        # Set up the mock human agent
        if self.config.mephisto.blueprint.get("onboarding_qualification", None):
            agent_id = self._register_mock_agents(num_agents=1, assume_onboarding=True)[
                0
            ]
        else:
            agent_id = self._register_mock_agents(num_agents=1)[0]

        # Set initial data
        self.await_channel_requests()

        # Make agent act
        self.server.submit_mock_unit(agent_id, task_data)
        self.await_channel_requests()

        return self.db.find_agents()[0].state.get_data()

    def _check_agent_state(
        self, state: Dict[str, Any], data_regression: DataRegressionFixture
    ):
        """
        Given an agent state, test that it is as expected.
        """
        data_regression.check(state)


class AbstractParlAIChatTest(AbstractCrowdsourcingTest):
    """
    Abstract class for end-to-end tests of one-turn ParlAIChatBlueprint tasks.
    """

    def _setup(self):
        super()._setup()
        self.message_sleep_time = 0  # Time to wait for any late messages to arrive

    def _test_agent_states(
        self,
        num_agents: int,
        agent_display_ids: Sequence[str],
        agent_messages: List[Sequence[str]],
        form_messages: Sequence[str],
        form_task_data: Sequence[Dict[str, Any]],
        expected_states: Sequence[Dict[str, Any]],
        agent_task_data: Optional[List[Sequence[Dict[str, Any]]]] = None,
    ):
        """
        Test that the actual agent states match the expected states.

        Register mock human agents, request initial data to define the 'inputs' fields
        of the agent states, make the agents have a conversation to define the 'outputs'
        fields of the agent states, and then check that the agent states all match the
        desired agent states.
        """

        # If no task data was supplied, create empty task data
        if agent_task_data is None:
            agent_task_data = []
            for message_round in agent_messages:
                agent_task_data.append([{}] * len(message_round))

        # Set up the mock human agents
        agent_ids = self._register_mock_agents(num_agents=num_agents)

        # # Feed messages to the agents

        # Have agents talk to each other
        assert len(agent_messages) == len(agent_task_data)
        for message_round, task_data_round in zip(agent_messages, agent_task_data):
            assert len(message_round) == len(task_data_round) == len(agent_ids)
            for agent_id, agent_display_id, message, task_data in zip(
                agent_ids, agent_display_ids, message_round, task_data_round
            ):
                self._send_agent_message(
                    agent_id=agent_id,
                    agent_display_id=agent_display_id,
                    text=message,
                    task_data=task_data,
                )

        # Have agents fill out the form
        for agent_idx, agent_id in enumerate(agent_ids):
            self.server.send_agent_act(
                agent_id=agent_id,
                act_content={
                    'text': form_messages[agent_idx],
                    'task_data': form_task_data[agent_idx],
                    'id': agent_display_ids[agent_idx],
                    'episode_done': False,
                },
            )
            self.await_channel_requests()

        # Submit the HIT
        for agent_id in agent_ids:
            self.server.submit_mock_unit(agent_id, {'final_data': {}})
            self.await_channel_requests()

        # # Check that the inputs and outputs are as expected

        # Get and filter actual messages
        time.sleep(self.message_sleep_time)
        actual_states = [agent.state.get_data() for agent in self.db.find_agents()]
        if len(actual_states) != len(expected_states):
            raise ValueError(
                f'There are {len(actual_states):d} agent states, instead of {len(expected_states):d} as expected!'
            )
        filtered_actual_states = []
        for actual_state in actual_states:
            filtered_actual_states.append(self._filter_agent_state_data(actual_state))

        # Check the contents of each message
        for actual_state, expected_state in zip(
            filtered_actual_states, expected_states
        ):
            clean_actual_state = self._remove_non_deterministic_keys(actual_state)
            assert clean_actual_state['inputs'] == expected_state['inputs']
            actual_num_messages = len(clean_actual_state['outputs']['messages'])
            expected_num_messages = len(expected_state['outputs']['messages'])
            if actual_num_messages != expected_num_messages:
                raise ValueError(
                    f'The actual number of messages is {actual_num_messages:d}, instead of {expected_num_messages:d} as expected!'
                )
            for actual_message, expected_message in zip(
                clean_actual_state['outputs']['messages'],
                expected_state['outputs']['messages'],
            ):
                for key, expected_value in expected_message.items():
                    self._check_output_key(
                        key=key,
                        actual_value=actual_message[key],
                        expected_value=expected_value,
                    )

    def _remove_non_deterministic_keys(self, actual_state: dict) -> dict:
        """
        Allow for subclasses to delete certain keys in the actual state that will change
        on each run.
        """
        return actual_state

    def _filter_agent_state_data(self, agent_state: dict) -> dict:
        """
        Remove agent state messages that do not contain text and are thus not useful for
        testing the crowdsourcing task.
        """
        filtered_messages = [
            m for m in agent_state['outputs']['messages'] if 'text' in m
        ]
        filtered_agent_state = {
            'inputs': agent_state['inputs'],
            'outputs': {**agent_state['outputs'], 'messages': filtered_messages},
        }
        return filtered_agent_state

    def _check_output_key(
        self: Union['AbstractParlAIChatTest', unittest.TestCase],
        key: str,
        actual_value: Any,
        expected_value: Any,
    ):
        # TODO: remove typing of self after switching to pytest regressions, in which we
        #  no longer inherit from TestCase
        """
        Check the actual and expected values, given that they come from the specified
        key of the output message dictionary.

        This function can be extended to handle special cases for subclassed Mephisto
        tasks.
        """
        if actual_value != expected_value:
            raise ValueError(
                f'The value of ["{key}"] is supposed to be {expected_value} but is '
                f'actually {actual_value}!'
            )

    def _send_agent_message(
        self, agent_id: str, agent_display_id: str, text: str, task_data: Dict[str, Any]
    ):
        """
        Have the agent specified by agent_id send the specified text and task data with
        the given display ID string.
        """
        act_content = {
            "text": text,
            "task_data": task_data,
            "id": agent_display_id,
            "episode_done": False,
        }
        self.server.send_agent_act(agent_id=agent_id, act_content=act_content)
        self.await_channel_requests()


def collapse_whitespace(in_string: str):
    """
    Helper function to remove extra whitespace that may make table outputs direct checks
    fail.
    """
    while "  " in in_string:
        in_string = in_string.replace("  ", " ")
    return in_string


def check_stdout(actual_stdout: str, expected_stdout_path: str):
    """
    Check that actual and expected stdouts match.

    Given a string of the actual stdout and a path to the expected stdout, check that
    both stdouts match, keeping in mind that the actual stdout may have additional
    strings relating to progress updates that are not found in the expected output
    strings.

    TODO: this can probably be moved to a method of an abstract test class once all
     analysis code relies on pytest regressions for some of its tests.
    """
    actual_stdout_lines = actual_stdout.split('\n')
    with open(expected_stdout_path) as f:
        expected_stdout = f.read()
    for expected_line in expected_stdout.split('\n'):
        if not any(
            collapse_whitespace(expected_line) in collapse_whitespace(actual_line)
            for actual_line in actual_stdout_lines
        ):
            raise ValueError(
                f'\n\tThe following line:\n\n{expected_line}\n\n\twas not found '
                f'in the actual stdout:\n\n{actual_stdout}'
            )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Generate knowledge grounded language modeling data.

Search a provided data source (given via a `--search-server` parameter)
for matching/similar context that may inform someone of how to generate
the next utterance for a given `--task`.

An `OverlapAgent` finds a matching "context" in all of <DATA SOURCE PROVIDED>
that is **distinct** from the provided context.

The goal is to find supporting information that doesn't come directly
from the document / context itself.

Usage:

```shell
python projects/seeker/scripts/generate_lm_data --task convai2 --search-server <relevant_server> \
    --save-dir /path/to/directory/for/log/saving
```

Upon building, one can then do the following to display the data:

```shell
$ parlai dd -t projects.seeker.tasks.lm:ResponseTeacher --root-dir /path/to/directory/for/log/saving/valid_split_0.5_f1_overlap
$ parlai dd -t projects.seeker.tasks.lm:KnowledgeTeacher --root-dir /path/to/directory/for/log/saving/valid_split_0.5_f1_overlap
$ parlai dd -t projects.seeker.tasks.lm:SearchQueryTeacher --root-dir /path/to/directory/for/log/saving/valid_split_0.5_f1_overlap
```
"""
import nltk
import os
import random
import torch
from typing import Optional, List, Tuple
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt

from parlai.agents.rag.retrievers import RetrieverType, Document
from parlai.agents.fid.fid import SearchQuerySearchEngineFiDAgent
from parlai.agents.tfidf_retriever.utils import filter_word
from parlai.core.agents import Agent, create_agent
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.metrics import F1Metric
from parlai.core.script import ParlaiScript
from parlai.core.worlds import create_task
from parlai.scripts.eval_model import setup_args, get_task_world_logs
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE
import parlai.utils.logging as logging
from parlai.utils.misc import TimeLogger
from parlai.utils.world_logging import WorldLogger
from projects.blenderbot2.agents.modules import BB2SearchQuerySearchEngineRetriever

INVALID = '<INVALID>'
DO_NOT_RETRIEVE = '<DO_NOT_RETRIEVE>'


class OverlapAgent(Agent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        SearchQuerySearchEngineFiDAgent.add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group('CC Overlap Group')
        group.add_argument(
            '--f1-overlap-threshold',
            type=float,
            default=0.5,
            help='Threshold of f1 in order to keep examples',
        )
        group.add_argument(
            '--min-num-search-words',
            type=float,
            default=-1,
            help='Threshold of how many words required in search term to search example (exclusive)',
        )
        group.add_argument(
            '--write-every-n-valid-exs',
            type=float,
            default=1000,
            help='How many valid exs to write for each log',
        )
        parser.set_defaults(rag_retriever_type=RetrieverType.SEARCH_ENGINE.value)
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt)
        if shared:
            self.search_engine = shared['search_engine']
        else:
            self.search_engine = OverlapSearchEngine(
                opt, DictionaryAgent(opt), shared=shared
            )  # type: ignore
        self.threshold = opt['f1_overlap_threshold']
        self.dummy = torch.zeros(1, 1, dtype=torch.long)

    def share(self):
        shared = super().share()
        shared['search_engine'] = self.search_engine
        return shared

    def construct_search_query(self, labels: List[str]) -> List[str]:
        """
        Construct the search query.

        :param observation:
            observation from task

        :return query:
            return search query.
        """
        assert labels
        search_query = [
            ' '.join(
                [
                    w
                    for w in DictionaryAgent.split_tokenize(labels[0])
                    if not filter_word(w)
                ]
            )
        ]
        return search_query

    def get_best_doc(
        self, all_docs: List[Document], labels: List[str]
    ) -> Tuple[Optional[float], Optional[Document], Optional[int]]:
        """
        Given a set of all retrieved docs, determine best fitting document.

        :param all_docs:
            list of all retrieved Documents
        :param labels:
            labels for the current example

        :return (best_f1, best_doc, best_doc_idx):
            return the best document, along with the f1 overlap and index into all_docs
        """
        docs = []
        for i, d in enumerate(all_docs):
            if d.startswith('.'):
                d = d[2:]
            try:
                docs += [(i, s) for s in nltk.sent_tokenize(d)]
            except IndexError:
                # Something's up with the NLTK Sentence tokenizer here.
                docs += [(i, s) for s in d.split('.')]
        f1s, inds = torch.FloatTensor(
            [F1Metric.compute(labels[0], [d]).value() for _, d in docs]
        ).topk(len(docs))
        best_doc = None
        best_doc_idx = None
        best_f1 = None
        for f1, ind in zip(f1s, inds):
            if self.threshold < f1 < 1.0 and labels[0] not in docs[ind][1]:
                best_doc = docs[ind][1]
                best_doc_idx = docs[ind][0]
                best_f1 = f1.item()
                break

        return best_f1, best_doc, best_doc_idx

    def act(self):
        """
        Search for overlap with the observation label.

        Return the best fitting document. A document is valid if the f1 is above the
        threshold AND the f1 is less than 1.0 AND the target label is not in the
        document.
        """
        obs = self.observation
        reply = {'text': INVALID, 'id': self.getID(), 'episode_done': False}
        if obs is None or obs['text'] == DO_NOT_RETRIEVE:
            return Message(reply)

        # construct the search query
        labels = obs.get('labels', obs.get('eval_labels', None))
        search_query = self.construct_search_query(labels)
        if (
            self.opt['min_num_search_words'] > 0
            and len(search_query[0].split()) <= self.opt['min_num_search_words']
        ):
            return Message(reply)

        # retrieve
        self.search_engine.set_search_queries(search_query)
        retrieved, _ = self.search_engine.retrieve_and_score(self.dummy)
        all_docs = [d.get_tokenization_str() for d in retrieved[0]]  # batched

        # Find the right doc
        best_f1, best_doc, best_doc_idx = self.get_best_doc(all_docs, labels)
        if best_doc:
            assert best_doc_idx is not None
            reply['knowledge'] = f'{TOKEN_KNOWLEDGE}{best_doc}{TOKEN_END_KNOWLEDGE}'
            reply['f1_overlap'] = best_f1
            reply['text'] = labels[0]
            reply['retrieved_docs'] = all_docs
            reply['gold_doc'] = all_docs[best_doc_idx]
            reply['search_query'] = search_query[0]
        return Message(reply)


class OverlapSearchEngine(BB2SearchQuerySearchEngineRetriever):
    def pick_chunk(self, query: str, doc_title: str, doc_text: str, doc_url: str):
        return [[doc_text]]


class GenerateLmData(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = setup_args()
        parser.add_argument(
            '--search-server',
            type=str,
            default=None,
            help='search server argument for retrieving documents',
        )
        parser.add_argument(
            '--save-dir', type=str, default=None, help='Path to directory to save data'
        )
        parser.set_params(model='projects.seeker.scripts.generate_lm_data:OverlapAgent')
        return parser

    def run(self):
        datatype = self.opt['datatype'].split(':')[0]
        self.opt['world_logs'] = os.path.join(
            self.opt['save_dir'],
            f"{datatype}_split_{self.opt['f1_overlap_threshold']}_f1_overlap",
        )
        try:
            self.generate_data()
        except:
            logging.error('ERROR')
            self.log()
            import ipdb

            ipdb.set_trace()
            raise

    def log(self):
        """
        Log world logs.
        """
        self.world_logger.reset()  # add final acts to logs
        outfile = self.task_opt['world_logs']
        os.makedirs(outfile, exist_ok=True)
        outfile = f'{outfile}/{self.cnt}_exs_log_part_{self.log_count}.jsonl'
        self.world_logger.write(
            outfile, self.world, file_format=self.opt['save_format']
        )
        self.world_logger._logs = []  # need to manually set this...
        self.log_count += 1

    def run_generation(self):
        """
        Actually run the evaluations.
        """
        # set up logging
        log_every_n_secs = self.opt.get('log_every_n_secs', -1)
        if log_every_n_secs <= 0:
            log_every_n_secs = float('inf')
        log_time = TimeLogger()

        # max number of examples to evaluate
        max_cnt = (
            self.opt['num_examples'] if self.opt['num_examples'] > 0 else float('inf')
        )
        self.cnt = 0
        self.n_valid = 0
        self.log_count = 0
        total_cnt = self.world.num_examples()

        while not self.world.epoch_done() and self.cnt < max_cnt:
            self.cnt += self.opt.get('batchsize', 1)
            self.world.parley()
            acts = self.world.get_acts()
            if acts[-1]['text'] != INVALID:
                try:
                    self.world.acts[0]['text'] += f"\n{acts[-1]['knowledge']}"
                except RuntimeError:
                    self.world.acts[0].force_set(
                        'text', f"{self.world.acts[0]['text']}\n{acts[-1]['knowledge']}"
                    )
                self.world.acts[0]['f1_overlap'] = acts[-1]['f1_overlap']
                self.world_logger.log(self.world)
                self.n_valid += 1
                if (
                    self.n_valid > 0
                    and self.n_valid % self.opt['write_every_n_valid_exs'] == 0
                ):
                    self.log()
            if log_time.time() > log_every_n_secs:
                report = self.world.report()
                report['n_valid'] = self.n_valid
                text, report = log_time.log(
                    report.get('exs', 0), min(max_cnt, total_cnt), report
                )
                logging.info(text)

    def generate_data(self):
        """
        Generate the LM Data.
        """
        random.seed(42)
        # load model and possibly print opt
        agent = create_agent(self.opt, requireModelExists=True)
        agent.opt.log()

        tasks = self.opt['task'].split(',')
        assert len(tasks) == 1
        task = tasks[0]
        logging.info(
            f'Generating data for task {task} using datatype {self.opt.get("datatype")}.'
        )
        logging.warning('Appending `flatten` to mutators.')
        # set up world logger
        self.task_opt = self.opt.copy()  # copy opt since we're editing the task
        self.task_opt['task'] = task
        if not self.task_opt['mutators']:
            self.task_opt['mutators'] = 'flatten'
        else:
            self.task_opt['mutators'] += '+flatten'
        # add task suffix in case of multi-tasking
        self.task_opt['world_logs'] = get_task_world_logs(
            task, self.task_opt['world_logs'], is_multitask=False
        )

        self.world_logger = WorldLogger(self.task_opt)

        self.world = create_task(self.task_opt, agent)  # create worlds for tasks

        self.run_generation()

        # dump world acts to file
        self.log()

        self.world.reset()


if __name__ == '__main__':
    GenerateLmData.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Tuple

import torch.jit


def is_alnum_or_underscore(ch: str):
    """
    A helper function for checking if a character is alphanumeric or underscore.
    """
    return ch.isalnum() or ch == '_'


@torch.jit.script
class ScriptableGpt2BpeHelper(object):
    """
    Version of parlai.utils.bpe.Gpt2BpeHelper that can be TorchScripted.
    """

    @classmethod
    def findall(cls, text: str) -> List[str]:
        """
        Split tokens in a manner that replicates parlai.utils.bpe.Gpt2BpeHelper.
        """
        contraction_endings = ["s", "t", "re", "ve", "m", "ll", "d"]

        tokens: List[str] = []
        idx = 0
        num_passes = 0
        while idx < len(text):
            num_passes += 1
            if num_passes > 10000:
                raise RuntimeError(
                    "*** Infinite loop in ScriptableGpt2BpeHelper.findall()! ***"
                )
            if text[idx] == "'":
                # Capture contradiction suffixes
                captured_suffix = False
                for ending in contraction_endings:
                    if text[idx + 1 : idx + 1 + len(ending)] == ending:
                        tokens.append("'" + ending)
                        idx += 1 + len(ending)
                        captured_suffix = True
                        break
                if captured_suffix:
                    continue
            if not text[idx].isspace() or (
                text[idx] == " " and idx + 1 < len(text) and not text[idx + 1].isspace()
            ):
                # Capture runs of one type of character
                if text[idx] == " ":
                    last_matching_idx = idx + 1
                else:
                    last_matching_idx = idx
                if text[last_matching_idx].isalpha():
                    while (
                        last_matching_idx + 1 < len(text)
                        and text[last_matching_idx + 1].isalpha()
                    ):
                        last_matching_idx += 1
                elif text[last_matching_idx].isnumeric():
                    while (
                        last_matching_idx + 1 < len(text)
                        and text[last_matching_idx + 1].isnumeric()
                    ):
                        last_matching_idx += 1
                else:
                    while (
                        last_matching_idx + 1 < len(text)
                        and not text[last_matching_idx + 1].isspace()
                        and not text[last_matching_idx + 1].isalpha()
                        and not text[last_matching_idx + 1].isnumeric()
                    ):
                        last_matching_idx += 1
                tokens.append(text[idx : last_matching_idx + 1])
                idx = last_matching_idx + 1
                continue
            if idx + 1 < len(text) and text[idx + 1].isspace():
                # Capture runs of space characters up until just before the final one
                last_space_idx = idx + 1
                while (
                    last_space_idx + 1 < len(text)
                    and text[last_space_idx + 1].isspace()
                ):
                    last_space_idx += 1
                if last_space_idx + 1 == len(text):
                    # Include the last char, which is a space char
                    tokens.append(text[idx : last_space_idx + 1])
                    idx = last_space_idx + 1
                else:
                    tokens.append(text[idx:last_space_idx])
                    idx = last_space_idx
                continue

            # Capture runs of space characters
            last_space_idx = idx
            while last_space_idx + 1 < len(text) and text[last_space_idx + 1].isspace():
                last_space_idx += 1
            tokens.append(text[idx : last_space_idx + 1])
            idx = last_space_idx + 1
        return tokens

    def __init__(
        self,
        add_prefix_space: bool,
        encoder: Dict[str, str],
        byte_encoder: Dict[int, str],
        fused_key_bpe_ranks: Dict[str, float],
        special_tokens: List[str],
    ):

        self.add_prefix_space = add_prefix_space

        self.encoder = encoder
        self.decoder: Dict[str, str] = {}
        if self.encoder is not None:
            for k, v in self.encoder.items():
                self.decoder[v] = k

        self.byte_encoder = byte_encoder
        self.byte_decoder: Dict[str, int] = {}
        if self.byte_encoder is not None:
            for k, v in self.byte_encoder.items():
                self.byte_decoder[v] = k

        self.bpe_ranks = fused_key_bpe_ranks

        # special tokens
        self._special_tokens: Dict[str, int] = {}
        for st in special_tokens:
            self._special_tokens[st] = 1

    def encode(self, text: str) -> List[str]:
        """
        Tokenize text.

        Checks for add_prefix_space; handles accordingly.

        :param text:
            text to tokenize

        :return tokens:
            A list of tokens
        """
        if self.add_prefix_space:
            text = f" {text}"

        # constants for readability
        FINAL = 1
        SPLITABLE = 0
        pieces: List[Tuple[str, int]] = [(text, SPLITABLE)]

        for special_token in self._special_tokens.keys():
            i = 0
            while i < len(pieces):
                subtext, status = pieces[i]
                if status == FINAL:
                    i += 1
                    continue
                split = subtext.split(special_token)
                if len(split) > 1:
                    # special token detected, replace the chunk with small subchunks
                    # split by the special token
                    pieces.pop(i)
                    for j, piece in enumerate(split):
                        if j > 0:
                            # add the special token as a delimiter
                            pieces.insert(i + (2 * j) - 1, (special_token, FINAL))
                        pieces.insert(i + (2 * j), (piece, SPLITABLE))
                else:
                    i += 1

        output: List[str] = []
        for piece, state in pieces:
            if state is FINAL:
                output.append(piece)
            else:
                output += self.helper_encode(piece)
        text = "".join(output)

        return output

    def get_pairs(self, word: List[str]) -> List[Tuple[str, str]]:
        """
        Return set of symbol pairs in a word.

        Word is represented as list of symbols (symbols being variable-length strings).

        :param word:
            word to symbolize

        :return pairs:
            set of tuples of symbols
        """
        pairs: List[Tuple[str, str]] = []
        prev_char = word[0]
        for char in word[1:]:
            pairs.append((prev_char, char))
            prev_char = char
        return pairs

    def bpe(self, word: List[str]) -> List[str]:
        """
        Convert token to BPE.

        :param word:
            list of tokens token to convert

        :return bpe_encoding:
            string bpe encoding
        """
        pairs = self.get_pairs(word)

        if len(pairs) == 0:
            return word

        while True:
            min_rank = self.bpe_ranks.get("\n".join(pairs[0]), float("inf"))
            bigram = pairs[0]
            for pair in pairs[1:]:
                current_rank = self.bpe_ranks.get("\n".join(pair), float("inf"))
                if current_rank < min_rank:
                    min_rank = current_rank
                    bigram = pair
            if "\n".join(bigram) not in self.bpe_ranks:
                break
            first, second = bigram
            new_word: List[str] = []
            i = 0
            while i < len(word):
                found = False
                for j in range(i, len(word)):
                    if word[j] == first:
                        new_word.extend(word[i:j])
                        i = j
                        found = True
                        break
                if not found:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            word = new_word.copy()
            if len(word) == 1:
                break
            else:
                pairs = self.get_pairs(word)
        return word

    def helper_encode(self, text: str) -> List[str]:
        """
        Tokenize text.

        :param text:
            text to tokenize

        :return tokens:
            A list of tokens
        """
        bpe_tokens: List[str] = []
        for token in self.findall(text):
            byte_encoded: List[str] = []
            for b in token:
                byte_encoded.append(self.byte_encoder[ord(b)])
            encoded: List[str] = []
            for bpe_token in self.bpe(byte_encoded):
                encoded.append(self.encoder[bpe_token])
            bpe_tokens.extend(encoded)
        return bpe_tokens

    def decode(self, tokens: List[str]) -> str:
        """
        Decode list of tokens into a text string.

        :param tokens:
            list of tokens

        :return text:
            decoded text
        """
        output: List[str] = []
        accum: List[str] = []
        for token in tokens:
            if token in self._special_tokens:
                if len(accum) > 0:
                    output.append(self.helper_decode(accum))
                    accum.clear()
                output.append(token)
            else:
                accum.append(token)
        if len(accum) > 0:
            output.append(self.helper_decode(accum))

        text = "".join(output)
        if self.add_prefix_space:
            assert text.startswith(" ")
            text = text.lstrip(" ")
        return text

    def helper_decode(self, tokens: List[str]) -> str:
        """
        Decode list of tokens into text string.

        :param tokens:
            list of tokens

        :return:
            decoded text
        """
        chars: List[str] = []
        for token in tokens:
            decoded_token = self.decoder[token]
            token_chars = self.utf8_chars(decoded_token)
            for char in token_chars:
                if not torch.jit.is_scripting():
                    # We iterate over "char", which is supposed to be a single
                    # character, because the TorchScripted version of the code
                    # correctly splits a string into single characters in
                    # self.utf8_chars() but the non-TorchScripted version doesn't
                    chars.extend(list(char))
                else:
                    chars.append(char)
        decoded_chars: List[str] = []
        for char in chars:
            decoded_chars.append(chr(self.byte_decoder[char]))
        return "".join(decoded_chars)

    def utf8_chars(self, s: str) -> List[str]:
        """
        An implementation of UTF8 character iteration in TorchScript. There are no
        bitwise operations in torchscript, so we compare directly to integer values.
        There isn't a lot of validation, for instance if you pass in an improperly
        encoded string with an out-of-place continuation byte, or with a non-left-to-
        right byte order, you'll get unexpected results and likely throw. Torch itself
        takes in unicode strings and encodes them as UTF8, so that should be actively
        hard to do.

        The logic is simple: looking at the current start-of-character byte.
        If its high bit is 0, it's a 1-byte character. Otherwise, the number of
        bytes is the number of leading 1s in its binary representation, so
        find that number by comparing it directly to ints with the appropriate
        representation, then append that many bytes as a character and move past
        them to the next start byte.

        From pytext.torchscript.utils.
        """
        chars: List[str] = []
        i = 0
        while i < len(s):
            byte = ord(s[i])
            if byte < 0b10000000:
                chars.append(s[i])
                i += 1
            else:
                if byte < 0b11100000:
                    num_bytes = 2
                elif byte < 0b11110000:
                    num_bytes = 3
                elif byte < 0b11111000:
                    num_bytes = 4
                elif byte < 0b11111100:
                    num_bytes = 5
                elif byte < 0b11111110:
                    num_bytes = 6
                elif byte < 0b11111111:
                    num_bytes = 7
                else:
                    num_bytes = 8
                chars.append(s[i : i + num_bytes])
                i += num_bytes
        return chars


@torch.jit.script
class ScriptableSubwordBpeHelper(object):
    """
    Version of parlai.utils.bpe.SubwordBPEHelper that can be TorchScripted.
    """

    @classmethod
    def findall(cls, text: str) -> List[str]:
        """
        Split tokens in a manner that replicates parlai.utils.bpe.Gpt2BpeHelper.
        """
        tokens: List[str] = []
        idx = 0
        num_passes = 0
        while idx < len(text):
            num_passes += 1
            if num_passes > 10000:
                raise RuntimeError(
                    "*** Infinite loop in ScriptableSubwordBpeHelper.findall()! ***"
                )

            if not text[idx].isspace():
                last_matching_idx = idx
                if is_alnum_or_underscore(text[last_matching_idx]):
                    while last_matching_idx + 1 < len(text) and (
                        text[last_matching_idx + 1].isalnum()
                        or text[last_matching_idx + 1] == '_'
                    ):
                        last_matching_idx += 1
                tokens.append(text[idx : last_matching_idx + 1])
                idx = last_matching_idx + 1
                continue

            # Capture runs of space characters
            last_space_idx = idx
            while last_space_idx + 1 < len(text) and text[last_space_idx + 1].isspace():
                last_space_idx += 1
            idx = last_space_idx + 1
        return tokens

    def __init__(
        self,
        add_prefix_space: bool,
        version: Tuple[int, int],
        bpe_codes: Dict[str, float],
        separator: str,
    ):

        self.add_prefix_space = add_prefix_space
        self.version = version
        self.bpe_codes = bpe_codes
        self.separator = separator

    def encode(self, text: str) -> List[str]:
        if self.add_prefix_space:
            text = f" {text}"
        return self.helper_encode(text)

    def get_pairs(self, word: List[str]) -> List[Tuple[str, str]]:
        """
        Return set of symbol pairs in a word.

        Word is represented as list of symbols (symbols being variable-length strings).

        :param word:
            word to symbolize

        :return pairs:
            set of tuples of symbols
        """
        pairs: List[Tuple[str, str]] = []
        prev_char = word[0]

        for char in word[1:]:
            pairs.append((prev_char, char))
            prev_char = char

        return pairs

    def bpe(self, orig: str) -> List[str]:
        """
        Convert token to BPE.

        :param word:
            token to convert

        :return bpe_encoding:
            string bpe encoding
        """
        if self.version == (0, 2):
            word = list(orig[:-1]) + [orig[-1] + '</w>']
        else:
            raise NotImplementedError

        pairs = self.get_pairs(word)
        if len(pairs) == 0:
            return [orig]

        while True:
            min_rank = self.bpe_codes.get("\n".join(pairs[0]), float("inf"))
            bigram = pairs[0]
            for pair in pairs[1:]:
                current_rank = self.bpe_codes.get("\n".join(pair), float("inf"))
                if current_rank < min_rank:
                    min_rank = current_rank
                    bigram = pair
            if "\n".join(bigram) not in self.bpe_codes:
                break
            first, second = bigram
            new_word: List[str] = []
            i = 0
            while i < len(word):
                found = False
                for j in range(i, len(word)):
                    if word[j] == first:
                        new_word.extend(word[i:j])
                        i = j
                        found = True
                        break
                if not found:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            word = new_word.copy()
            if len(word) == 1:
                break
            else:
                pairs = self.get_pairs(word)

        # don't print end-of-word symbols
        if word[-1] == '</w>':
            word = word[:-1]
        elif word[-1].endswith('</w>'):
            word = word[:-1] + [word[-1].replace('</w>', '')]
        return word

    def segment_tokens(self, tokens: List[str]) -> List[str]:
        """
        segment a sequence of tokens with BPE encoding.
        """
        output: List[str] = []
        for word in tokens:
            # eliminate double spaces
            if not word:
                continue

            new_word: List[str] = []
            for out in self.bpe(word):
                new_word.append(out)

            for item in new_word[:-1]:
                output.append(item + self.separator)
            output.append(new_word[-1])

        return output

    def helper_encode(self, text: str) -> List[str]:
        text = text.replace('\n', ' __newln__ ')
        tokens = self.findall(text)
        return self.segment_tokens(tokens)

    def decode(self, tokens: List[str]) -> str:
        output: List[str] = []
        accum: List[str] = []
        for token in tokens:
            accum.append(token)
        if len(accum) > 0:
            output.append(self.helper_decode(accum, ' '))

        text = "".join(output)
        if self.add_prefix_space:
            assert text.startswith(" ")
            text = text.lstrip(" ")
        return text

    def helper_decode(self, tokens: List[str], delimiter: str = ' ') -> str:
        text = delimiter.join(tokens)
        text = text.replace('@@ ', '')
        # It's also possible that we get a BPE encoding on the end of the word
        if text.endswith('@@'):
            text = text[:-2]
        text = text.replace('__newln__', '\n')
        return text


@torch.jit.script
class ScriptableDictionaryAgent:
    """
    Builds and/or loads a dictionary.

    All code is TorchScriptable.
    """

    def __init__(
        self,
        null_token: str,
        end_token: str,
        unk_token: str,
        start_token: str,
        freq: Dict[str, int],
        tok2ind: Dict[str, int],
        ind2tok: Dict[int, str],
        bpe_add_prefix_space: bool,
        bpe_encoder: Dict[str, str],
        bpe_byte_encoder: Dict[int, str],
        fused_key_bpe_ranks: Dict[str, float],
        special_tokens: List[str],
        subword_bpe_version: Tuple[int, int],
        fused_bpe_codes: Dict[str, float],
        subword_bpe_separator: str,
    ):

        self.null_token = null_token
        self.end_token = end_token
        self.unk_token = unk_token
        self.start_token = start_token

        self.freq = freq
        self.tok2ind = tok2ind
        self.ind2tok = ind2tok

        # cache unk token for later
        self._unk_token_idx = self.tok2ind[self.unk_token]

        # Initialize tokenizer
        self.subword_bpe = ScriptableSubwordBpeHelper(
            add_prefix_space=bpe_add_prefix_space,
            version=subword_bpe_version,
            bpe_codes=fused_bpe_codes,
            separator=subword_bpe_separator,
        )

        self.gpt2_bpe = ScriptableGpt2BpeHelper(
            add_prefix_space=bpe_add_prefix_space,
            encoder=bpe_encoder,
            byte_encoder=bpe_byte_encoder,
            fused_key_bpe_ranks=fused_key_bpe_ranks,
            special_tokens=special_tokens,
        )

    def _word_lookup(self, key: str) -> int:
        """
        Return index from token, or unk_token's index, or None.
        """
        if key in self.tok2ind:
            return self.tok2ind[key]
        else:
            return self._unk_token_idx

    def _index_lookup(self, key: int) -> str:
        """
        Return token from index, or unk_token.
        """
        if key in self.ind2tok:
            return self.ind2tok[key]
        else:
            return self.unk_token

    def gpt2_tokenize(self, text: str):
        """
        Tokenize using Gpt2 BPE tokenizer.
        """
        return self.gpt2_bpe.encode(text)

    def bpe_tokenize(self, text: str) -> List[str]:
        """
        Return a sequence of BPE-tokens from the text.
        """
        return self.subword_bpe.encode(text)

    def txt2vec(self, text: str, dict_tokenizer: str) -> List[int]:
        """
        Convert a string to a vector (list of ints).

        First runs a sentence tokenizer, then a word tokenizer.
        """
        itr: List[int] = []
        if dict_tokenizer == 'gpt2' or dict_tokenizer == 'slow_bytelevel_bpe':
            for token in self.gpt2_tokenize(str(text)):
                itr.append(self._word_lookup(token))
        elif dict_tokenizer == 'bpe':
            for token in self.bpe_tokenize(str(text)):
                itr.append(self._word_lookup(token))
        else:
            raise NotImplementedError
        return itr

    def vec2txt(self, vector: List[int], dict_tokenizer: str) -> str:
        """
        Convert a vector of IDs to a string.

        Converts a vector (iterable of ints) into a string, with each token separated by
        the delimiter (default ``' '``).
        """
        tokens = [self._index_lookup(idx) for idx in vector]
        if dict_tokenizer == 'gpt2' or dict_tokenizer == 'slow_bytelevel_bpe':
            text = self.gpt2_bpe.decode(tokens)
        elif dict_tokenizer == 'bpe':
            text = self.subword_bpe.decode(tokens)
        else:
            raise NotImplementedError
        return text



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.tasks.google_sgd.build as original_google_sgd_build
import parlai.core.build_data as build_data

import os
import json
import random
from shutil import copyfile

DATA_LEN = original_google_sgd_build.DATA_LEN

MODEL_MODEL_HOLDOUT_DOMAINS = [
    "Homes_1",
    "Homes_2",
    "FindHomeByArea",
    "RentalCars_1",
    "RentalCars_2",
    "RentalCars_3",
    "Messaging_1",
    "Payment_1",
]


def build(opt):
    # get path to data directory
    dpath = os.path.join(opt["datapath"], "google_sgd_rl_splits")
    # define version if any
    version = "1.0"

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # Grab things from the original Google SGD
        original_google_sgd_build.build(opt)

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        model_model_convos = []
        model_model_schemas = {}
        tot_count = 0
        for split_type in ["train", "dev", "test"]:
            outpath = os.path.join(dpath, split_type)
            os.makedirs(outpath, exist_ok=True)
            original_path = os.path.join(opt["datapath"], "google_sgd", split_type)

            copyfile(
                os.path.join(original_path, "schema.json"),
                os.path.join(outpath, "schema.json"),
            )
            with open(os.path.join(original_path, "schema.json")) as f:
                schemas = json.load(f)
            for schema in schemas:
                model_model_schemas[schema["service_name"]] = schema

            for file_id in range(1, DATA_LEN[split_type] + 1):
                filename = f"dialogues_{file_id:03d}.json"
                original_file = os.path.join(original_path, filename)
                with open(original_file) as f:
                    blobs = json.load(f)
                save_data = []
                for blob in blobs:
                    blob[
                        "dialogue_id"
                    ] = f"{blob['dialogue_id']}_from_{split_type}_{file_id:03d}"
                    in_model_model = False
                    for service in blob["services"]:
                        if service in MODEL_MODEL_HOLDOUT_DOMAINS:
                            in_model_model = True
                            tot_count += 1
                    if in_model_model:
                        model_model_convos.append(blob)
                    else:
                        save_data.append(blob)
                with open(os.path.join(outpath, filename), "w+") as f:
                    json.dump(save_data, f, indent=4)
                print(split_type, filename)

        print("done processing train + dev + test ")
        # deal with custom splits
        print(
            "number of samples in homes + rental cars + messaging + payments",
            len(model_model_convos),
        )
        print("service usage count of above domains", tot_count)
        model_model_path = os.path.join(dpath, "model_model_splits")
        os.makedirs(model_model_path, exist_ok=True)
        random.Random(42).shuffle(model_model_convos)

        def save_model_model(convos, split_type, model_model_path, schema):
            os.makedirs(os.path.join(model_model_path, split_type), exist_ok=True)
            for i in range(int(len(convos) / 64) + 1):
                with open(
                    os.path.join(
                        model_model_path, split_type, f"dialogues_{i:03d}.json"
                    ),
                    "w+",
                ) as f:
                    json.dump(convos[i * 64 : (i + 1) * 64], f, indent=4)
            with open(
                os.path.join(model_model_path, split_type, "schema.json"), "w+"
            ) as f:
                json.dump(list(schema.values()), f, indent=4)

        save_model_model(
            model_model_convos[: int(0.6 * len(model_model_convos))],
            "train",
            model_model_path,
            model_model_schemas,
        )
        save_model_model(
            model_model_convos[
                int(0.6 * len(model_model_convos)) : int(0.8 * len(model_model_convos))
            ],
            "dev",
            model_model_path,
            model_model_schemas,
        )
        save_model_model(
            model_model_convos[int(0.8 * len(model_model_convos)) :],
            "test",
            model_model_path,
            model_model_schemas,
        )

        print("done processing test")

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)
